/**
 *                      ==================
 *                      IITK OSD Simulator
 *                      ==================
 *   An Implementation of Simulator for Object based storage device(OSD)
 *
 *   Copyright (C) 2007-2008 CSE IIT KANPUR
 *
 *   Contact: Department of Computer Science and Engineering
 *            Indian Institute of Technology,Kanpur(IIT Kanpur)
 *            {moona,riyaz}@cse.iitk.ac.in
 *
 *   This file is part of IITK OSD Simulator Project.
 *
 *   IITK OSD Simulator Project is free software; you can redistribute it
 *   and/or modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   IITK OSD Simulator Project is distributed in the hope that it will be
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the  GNU General
 *   Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

# include <strings.h>
# include <stdio.h>
# include <unistd.h>
# include <stdlib.h>
# include <string.h>
# include<sys/types.h>
# include<sys/socket.h>
# include<netinet/in.h>
# include<arpa/inet.h>
# include<errno.h>

# include "osd_req_ops.h"
# include "osd_export.h"

/**
**********************************
 * File:  osd-sim.c
 * - Create OSD server port for listening to command processing request
 * - Accept connection request from iSCSI Module and spwans process for each   
 *   request
 * - Calls request processing functions for the OSD request
 **/

/** Server TCP port no to listen for the request */
int SERV_TCP_PORT=11026;

/** Listening on local machine */
#define SERV_HOST_ADDR "127.0.0.1"

/** Buffer used to send acknowledgements */
#define HELLO_BUF_SIZE 100

/** Debug level */
int DEBUG=-1;

int dlevel=UPPER_FUNC;
/** Structure to maintain list of all OSD Luns in the system */
LUN_STRUCT *lun_struct=NULL;

/**OSD directory for the request */
char osd_directory[FILENAME_SIZE]="/tmp/osd";

/** OSD configuration file name */
char osd_conf_file[FILENAME_SIZE]="/etc/osd_export.conf"; 
char iscsi_conf_file[FILENAME_SIZE]="/etc/ietd.conf"; 



void usage(void) {
    printf("\n");
    printf("\n             ** IITK OSD SIMULATOR PROGRAM **                      \n");
    printf("OSD Simulator Accepts OSD Request From iSCSI Target And Processes It \n\n");
    printf("Usage: -p <port>           Server Port No (>1024) to Listen OSD Request\n");
    printf("			       		Default Port: 11026 \n");		
    printf("       -d <debug_level>    Debug level for OSD simulator\n");
    printf("					Default Level: -1\n"); 
    printf("					-1 : Run OSD Simulator as Background Deamon \n");
    printf("					 0 : Error Messages\n");	
    printf("					 1 : Upper level function messages\n");	
    printf("					 2 : Detailed function messages\n");		
    printf("       -o<filename>        Name of OSD configuration file to export OSD Luns\n");
    printf("					Default file: /etc/osd_export.conf\n");
    printf("       -i<filename>        Name of iSCSI configuration file to export iSCSI Luns\n");
    printf("					Default file: /etc/ietd.conf\n");
    printf("       -h/--help           To Print Usage information\n");
    printf("\n");
}

int main(int argc,char *argv[])
{

	int sockfd,newsockfd,childpid,i;
	socklen_t clilen;
	struct sockaddr_in cli_addr,serv_addr;
		
	if(argc>1)
	{
		for (i=1; i<argc; i++) 
		{
			if (!strcmp(argv[i], "-p")) {
				i++;
				if(i<argc)
				{
					SERV_TCP_PORT = atoi(argv[i]); 
		
				}
				else
				{
					printf("Enter Port No value after -p ");
					usage();
					exit(0);	
				}	
			} else if (!strcmp(argv[i], "-d")) 
			{
				i++;
				if(i<argc)
				{
					DEBUG = atoi(argv[i]); 
				}
				else
				{
					printf("Enter DEBUG level after -d ");
					usage();	
					exit(0);	
				}	
			} else if (!strcmp(argv[i], "-o")) 
			{
				i++;
				if(i<argc)
				{
					strcpy(osd_conf_file,argv[i]); 
					
				}
				else
				{
					printf("Enter OSD Configuration File after -o\n");
					usage();
					exit(0);	
				}	
			} else if (!strcmp(argv[i], "-i")) 
			{
				i++;
				if(i<argc)
				{
					strcpy(iscsi_conf_file,argv[i]); 
					
				}
				else
				{
					printf("Enter iSCSI Configuration File after -i\n");
					usage();
					exit(0);
				}	
			} else if(!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help") )
			{
				usage();
				exit(0);	
			} else
			{
				printf(" Wrong Configuration option\n");
				usage();
				exit(0);
			}
		
		}
	
	}	

	
	if(DEBUG ==-1)
       {
	   /** Start OSD simulator as background deamon process*/ 
		int pid = fork();
		if (pid < 0) 
		{
			print_msg(ERROR_MSG,"Starting OSD Deamon failed");
			exit(1);
		} else if (pid)
			exit(0);
		chdir("/");
	}
	printf("\n IITK OSD Simulator Program Started \n");
	
 	/** Create List of OSD Luns in the System */
	lun_struct=prepare_osd_luns(osd_conf_file);
	
	if(lun_struct==NULL)
	{
		print_msg(ERROR_MSG,"Error in prepare_osd_luns");
		return 1;
	}	
	
	/** Export OSD Luns as iSCSI Luns*/
	if(export_osd_luns(lun_struct,iscsi_conf_file))
	{
		print_msg(ERROR_MSG,"Error in export_osd_luns");
		return 1;
	}
	
	/**Initialize page related structure variables */ 
	//Never forget to call this
	initialize_all_pages_struct();
	
	
	/** Create Socket Server to listen for OSD request from iSCSI Module */
	if((sockfd=socket(AF_INET,SOCK_STREAM,0))<0)
        {	print_msg(ERROR_MSG,"server error");
                return 0;
        }
	bzero((char *) &serv_addr,sizeof(serv_addr));          
	serv_addr.sin_family=AF_INET;
	serv_addr.sin_addr.s_addr=htonl(INADDR_ANY);
	serv_addr.sin_port=htons(SERV_TCP_PORT);
	if(bind(sockfd,(struct sockaddr *)&serv_addr,sizeof(serv_addr))<0)
	{	print_msg(ERROR_MSG,"server bind error");
		return 0;
        }
	listen(sockfd,5);
     	for(; ;)	
	{
		clilen=sizeof(cli_addr);
		newsockfd=accept(sockfd,(struct sockaddr *)&cli_addr,&clilen); 
		if(newsockfd<0)
		{	print_msg(ERROR_MSG,"server accept error");
			return 1;
		}	
		/** Fork process for request */
                if((childpid=fork())<0)
                {
                        print_msg(ERROR_MSG,"server fork error");
                	return 1;
                }
              	else if(childpid==0)
                {
                     close(sockfd);
                     osd_req_process(newsockfd);
                     exit(0);
                }

               close(newsockfd);   
	}

	return 0;	
}

/** Process OSD request */
int osd_req_process(int sockfd)
{	
	print_msg(dlevel,"Request accepted, processing now\n");
	OSD_REQUEST *osd_req;
	

	
	
	/** Allocate OSD Request structure*/
	osd_req=(OSD_REQUEST *)allocate_osd_request();
	/** Initialize request processing status bits*/
	init_not_initiated_command_functions_bit(osd_req->not_initiated_command_functions,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	
	/** Receive OSD Reqest */
	if(osd_request_receive(osd_req,sockfd))
	{
		print_msg(ERROR_MSG,"Error in receiving osd_request");
		return 0;
	}
	
	/** Extract OSD Lun from request and set to osd_req LUN */ 
	if(set_osd_req_lun(osd_req,lun_struct))
	{
		print_msg(ERROR_MSG,"Error in set_osd_req_lun");
		return 0;
	}
	
	/** OSD Base Directory */
	strcpy(osd_directory,osd_req->osd_lun->osd_lun_base_directory);
	

	/** Process OSD request */
	if(osd_request_process(osd_req))
	{
		print_msg(ERROR_MSG,"Error in osd_request_process");
		goto send_response;		
	}
	
	/** Extrace Get/Set Attributes*/
	if(osd_request_extract_get_set_attr(osd_req))
	{
		print_msg(ERROR_MSG,"Error in osd_request_extract_get_set_attr");
		goto send_response;		
	}
	
	/** Process Get/Set attributes*/
	if(((OSD_CDB_ATTR_LIST_HDR *)(&(osd_req->osd_cdb_struct)))->set_list_length)
	{
	if(process_set_attribute_list(osd_req))
	{
		print_msg(ERROR_MSG,"Error in process_set_attribute_list");
		goto send_response;		
	}

	}
	//get-set attr implement
	if(((OSD_CDB_ATTR_LIST_HDR *)(&(osd_req->osd_cdb_struct)))->get_list_length)
	{
		if(process_get_attribute_list(osd_req))
		{
			print_msg(ERROR_MSG,"Error in process_get_attribute_list");
			goto send_response;		
		}

	}
	
	
	/** Send OSD command response */
send_response:		
	if(osd_request_send_response(osd_req,sockfd))
	{
		print_msg(ERROR_MSG,"Error in osd_request_send_response");
	
	
	}
	free_osd_request(osd_req);
		
	exit(0);
	
}
