/**
 *                      ==================
 *                      IITK OSD Simulator
 *                      ==================
 *   An Implementation of Simulator for Object based storage device(OSD)
 *
 *   Copyright (C) 2007-2008 CSE IIT KANPUR
 *
 *   Contact: Department of Computer Science and Engineering
 *            Indian Institute of Technology,Kanpur(IIT Kanpur)
 *            {moona,riyaz}@cse.iitk.ac.in
 *
 *   This file is part of IITK OSD Simulator Project.
 *
 *   IITK OSD Simulator Project is free software; you can redistribute it
 *   and/or modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   IITK OSD Simulator Project is distributed in the hope that it will be
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the  GNU General
 *   Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef OSD_ATTR2_H
#define OSD_ATTR2_H
#include "osd_attr_manip.h"


/**
**********************************
 * File:  osd_attr2.h
 * Functions that implements changes in attribute pages as side effect of OSD command * 
 * execution
 **/



/** Error code values used internally */
# define ATTR_UPDATE_OK 0
# define UOBJ_LENGTH_QUOTA_EXCEED 2
# define UOBJ_CAPACITY_QUOTA_EXCEED 3 
# define PART_NO_USER_OBJ_QUOTA_EXCEED 4
# define PART_CAPACITY_QUOTA_EXCEED 5
# define ROOT_CAPACITY_QUOTA_EXCEED 6
# define ROOT_NO_PART_OBJ_QUOTA_EXCEED 7
# define INTERNAL_PAGE_CREATE_ERROR 1


/**
 *
 *    Function:    osd_format_attribute_pages_effects
 *
 *    Description:  Implements side effect of OSD FORMAT command
 *			- Create Root attribute pages 
 *			- Creates partition 0 attribute pages
 *		
 * 
 *    Arguments:    @osd_system_id - OSD system Identifier
 *		    @capacity - Capacity of OSD Lun	 
 *		    @username - User Name for OSD Lun
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/


int osd_format_attribute_pages_effects(const char *osd_system_id,uint64_t capacity,
					const char *username);


/**
 *
 *    Function:    create_partition_attribute_pages_effects
 *
 *    Description:  Implements side effect of OSD CREATE PARTITION  command
 *		    Creates partition  attribute pages
 *		
 * 
 *    Arguments:    @part_id - partition ID
 *		    @username - User Name for OSD Lun
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

int create_partition_attribute_pages_effects(uint64_t part_id,const char *username);


/**
 *
 *    Function:    create_uobj_coll_attribute_pages_effects
 *
 *    Description:  Implements side effect of OSD CREATE_OBJECT/CREATE_COLLECTION command
		    Creates User Object /Collection Object attribute pages
		
 * 
 *    Arguments:    @part_id - partition ID
 *		    @object_id - user /collection object id
 *		    @object_type - Type of Object
 *				UOBJ_OBJECT_TYPE 0x80 
 *				COLL_OBJECT_TYPE 0x40					
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

int create_uobj_coll_attribute_pages_effects(uint64_t part_id,uint64_t object_id,
					     int object_type);


/**
 *
 *    Function:    add_to_coll_attribute_pages_effects
 *
 *    Description:  Implements side effect of when User object is added to collection list 
 *
 * 
 *    Arguments:    @part_id - partition ID
 *		    @coll_object_id - Collection object		
 *		    @object_id - user object id added
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

int add_to_coll_attribute_pages_effects(uint64_t part_id,uint64_t coll_object_id,
				        int object_id);

/**
 *
 *    Function:    remove_from_coll_attribute_pages_effects
 *
 *    Description:  Implements side effect of when User object is removed from collection
 * 		     list 
 *
 * 
 *    Arguments:    @part_id - partition ID
 *		    @coll_object_id - Collection object		
 *		    @object_id - user object id removed
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

int remove_from_coll_attribute_pages_effects(uint64_t part_id,uint64_t coll_object_id,
					    int object_id);



/**
 *
 *    Function:    read_attribute_pages_effects
 *
 *    Description:  Implements side effect of OSD READ command
 *		
 * 
 *    Arguments:    @part_id - partition ID
 *		    @object_id - User Object ID
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/
int read_attribute_pages_effects(uint64_t part_id,uint64_t object_id);

/**
 *
 *    Function:    list_attribute_pages_effects
 *
 *    Description:  Implements side effect of OSD LIST command
 *		
 * 
 *    Arguments:    @part_id - partition ID
 *		    @coll_obj_id - Collection Object ID
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

int list_attribute_pages_effects(uint64_t part_id,uint64_t coll_object_id);


/**
 *
 *    Functions:    getattr_attribute_pages_effects
 *		    setattr_attribute_pages_effects
 *
 *    Description:  Implements side effect of OSD GETATTR and SETATTR command
 *		
 * 
 *    Arguments:    @part_id - partition ID
 *		    @object_id - Object ID
 * 		    @object_type - Type of object 		
 *					UOBJ_OBJECT_TYPE 0x80 
 *					COLL_OBJECT_TYPE 0x40
 *					PART_OBJECT_TYPE 0x2 
 *					ROOT_OBJECT_TYPE 0x1
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

int getattr_attribute_pages_effects(uint64_t part_id,uint64_t object_id,int object_type);
int setattr_attribute_pages_effects(uint64_t part_id,uint64_t object_id,int object_type);


/**
 *
 *    Functions:    remove_uobj_coll_attribute_pages_effects
 *
 *    Description:  Implements side effect of OSD REMOVE and REMOVE_COLLECTION command
 *		
 * 
 *    Arguments:    @part_id - partition ID
 *		    @object_id - Object ID
 * 		    @object_type - Type of object 		
 *					UOBJ_OBJECT_TYPE 0x80 
 *					COLL_OBJECT_TYPE 0x40
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/
int remove_uobj_coll_attribute_pages_effects(uint64_t part_id,uint64_t object_id,
					     int object_type);

/**
 *
 *    Function:    remove_partition_attribute_pages_effects
 *
 *    Description:  Implements side effect of OSD REMOVE_PARTITION command
 *		
 * 
 *    Argument:    @part_id - partition ID
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

int remove_partition_attribute_pages_effects(uint64_t part_id);




/** Utility structure used while checking quota voilance and attribute 
   modification for OSD_WRITE,OSD_APPEND,OSD_CREATE_N_WR command
**/ 
typedef struct write_afftected_attr_struct
{
 uint64_t object_id;
 uint64_t part_id;
 uint64_t root_used_capacity;
 //uint64_t root_total_capacity;
 uint64_t part_used_capacity;
 //uint64_t part_capacity_quota;
 uint64_t user_obj_logical_len;
 //uint64_t user_max_obj_len;
 uint64_t user_used_capacity;
 uint64_t length;
 uint64_t starting_byte_offset;
}WRITE_AFFECTED_ATTRS;

/**
 *
 *    Function:    get_and_check_write_affected_attributes
 *
 *    Description: Object_id and oart_id of WRITE_AFFECTED_ATTRS are set by caller 
 * 		   Function checks for quota voilation and sets following fields of  	
 *		   WRITE_AFFECTED_ATTRS  	 
 *		    -  root_used_capacity, part_used_capacity,
 *		    -  user_obj_logical_len,user_used_capacity,
 *
 *    Argument:    @service_action - Can be OSD_WRITE,OSD_APPEND,OSD_CREATE_N_WR
 *		   @write_affected_attrs - fields are set in this function	
 * 
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/
int get_and_check_write_affected_attributes(uint16_t service_action,WRITE_AFFECTED_ATTRS *write_affected_attrs);


/**
 *
 *    Function:    set_write_affected_attributes
 *
 *    Description: Function modifies attribute pages according to WRITE_AFFECTED_ATTRS 
 * 		   fields
 *	 
 *    Argument:  @write_affected_attrs - attributes are updated according to fields 
 * 
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

int set_write_affected_attributes(WRITE_AFFECTED_ATTRS *write_affected_attrs);

#define OSD_CMD_APPEND            0x8807
#define OSD_CMD_CREATE_AND_WRITE  0x8812
#define OSD_CMD_WRITE             0x8806

//int create_write_uobj_coll_attribute_pages_effects(uint64_t part_id,uint64_t object_id,const char *username,int object_type);

#endif /* ifdef OSD_ATTR2_H*/

