/**
 *                      ==================
 *                      IITK OSD Simulator
 *                      ==================
 *   An Implementation of Simulator for Object based storage device(OSD)
 *
 *   Copyright (C) 2007-2008 CSE IIT KANPUR
 *
 *   Contact: Department of Computer Science and Engineering
 *            Indian Institute of Technology,Kanpur(IIT Kanpur)
 *            {moona,riyaz}@cse.iitk.ac.in
 *
 *   This file is part of IITK OSD Simulator Project.
 *
 *   IITK OSD Simulator Project is free software; you can redistribute it
 *   and/or modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   IITK OSD Simulator Project is distributed in the hope that it will be
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the  GNU General
 *   Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef OSD_ATTR_MANIP_H
#define OSD_ATTR_MANIP_H
# include "osd_attr1.h"
/**
**********************************
 * File:  osd_attr_manip.h
 * Macros and Function declarations related to get/set attribute
**/

/** OSD Object type code according to T10 standard */
#define UOBJ_OBJECT_TYPE 0x80 
#define COLL_OBJECT_TYPE 0x40
#define PART_OBJECT_TYPE 0x2 
#define ROOT_OBJECT_TYPE 0x1


/** Macro For get/set attribute for  Current command page */
#define CURR_CMD_OBJECT_TYPE 0x0 


/**
 *
 *    Functions:    set_page_attr
 * 		    get_page_attr
 *	
 *
 *    Description: Get /Set page attribute 
 *
 *    Argument:  @part_id - Partition ID
 *		 @object_id -Object ID
 *		 @page_no - Page No
 *		 @attr_no - Attribute no
 *		 @attr_value - Pointer to attribute value buffer
 *		 @object_type - Type of Object	
 *		 			UOBJ_OBJECT_TYPE 0x80 
 *					COLL_OBJECT_TYPE 0x40
 *					PART_OBJECT_TYPE 0x2 
 *					ROOT_OBJECT_TYPE 0x1
 *				
 * 
 *    Return codes: 0 if success, 1  otherwise 
 *
 **/
int  set_page_attr(uint64_t part_id,uint64_t object_id, uint32_t page_no,uint32_t attr_no,void *attr_value,int object_type); 
int get_page_attr(uint64_t part_id,uint64_t  object_id, uint32_t page_no,uint32_t attr_no,void *attr_value,int object_type); 


/**
 *
 *    Functions:    get_osd_attr_struct
 *
 *    Description: Functions returns OSD_ATTR_INFO(defiend in osd_attr.h)
 *		   for the given object which gives information about attribute size,its 
 *		   offset in attribute file,settable/not.This function is called by 
 *		   get/set_page_attr  
 *  	
 *
 *    Argument:  @part_id - Partition ID
 *		 @object_id -Object ID
 *		 @page_no - Page No
 *		 @attr_no - Attribute no
 *		 @object_type - Type of Object	
 *		 			UOBJ_OBJECT_TYPE 0x80 
 *					COLL_OBJECT_TYPE 0x40
 *					PART_OBJECT_TYPE 0x2 
 *					ROOT_OBJECT_TYPE 0x1
 *				
 * 
 *    Return code: OSD_ATTR_INFO with fields set for attribute size,attribute offset in 
 * 		    file and settable flag.If attribute/page not found sets attribute
 * 		    offset to OSD_ATTR_WRONG_OFFSET which is condered as error.   
 *
 **/

OSD_ATTR_INFO get_osd_attr_struct(uint64_t part_id,uint64_t object_id, uint32_t page_no,uint32_t attr_no,int object_type);



/**
 *
 *    Functions:    find_offset_root_page
 * 		    find_offset_partition_page
 *		    find_offset_uobject_page
 *		    find_offset_collection_page	 
 *		   find_offset_current_command_page		
 *
 *    Description: Find offset of given attribute in Object page associated with object of *		   given type
 *
 *    Argument:  @page_no - Page No
 *		 @attr_no - Attribute no
 * 
 *    Return codes: OSD_ATTR_INFO with fields set for attribute size,attribute offset in 
 * 		    file and settable flag.If attribute/page not found sets attribute
 * 		    offset to OSD_ATTR_WRONG_OFFSET which is condered as error.   
 *
 **/

OSD_ATTR_INFO find_offset_root_page( uint32_t page_no,uint32_t attr_no);
OSD_ATTR_INFO find_offset_partition_page( uint32_t page_no,uint32_t attr_no);
OSD_ATTR_INFO find_offset_uobject_page( uint32_t page_no,uint32_t attr_no);
OSD_ATTR_INFO find_offset_collection_page( uint32_t page_no,uint32_t attr_no);
OSD_ATTR_INFO find_offset_current_command_page( uint32_t page_no,uint32_t attr_no);



/**
 *
 *    Functions:  get_offset_of_page_in_file
 *
 *    Description: Returns offset of given page in attribute file associated with given 
* 		   object type
 *
 *    Argument:  @part_id - Partition ID
 *		 @object_id -Object ID
 *		 @page_no - Page No 
 *		 @object_type - Type of Object	
 *		 			UOBJ_OBJECT_TYPE 0x80 
 *					COLL_OBJECT_TYPE 0x40
 *					PART_OBJECT_TYPE 0x2 
 *					ROOT_OBJECT_TYPE 0x1
 *				
 * 
 *    Return codes: actual offset if success, 0xFFFF otherwise 
 *
 **/
uint16_t get_offset_of_page_in_file( uint64_t part_id,uint64_t object_id, uint32_t page_no,int object_type);

#endif/* ifdef OSD_ATTR_MANIP_H */
