/**
 *                      ==================
 *                      IITK OSD Simulator
 *                      ==================
 *   An Implementation of Simulator for Object based storage device(OSD)
 *
 *   Copyright (C) 2007-2008 CSE IIT KANPUR
 *
 *   Contact: Department of Computer Science and Engineering
 *            Indian Institute of Technology,Kanpur(IIT Kanpur)
 *            {moona,riyaz}@cse.iitk.ac.in
 *
 *   This file is part of IITK OSD Simulator Project.
 *
 *   IITK OSD Simulator Project is free software; you can redistribute it
 *   and/or modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   IITK OSD Simulator Project is distributed in the hope that it will be
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the  GNU General
 *   Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef OSD_FILE_MANIP_H
#define OSD_FILE_MANIP_H

/**
**********************************
 * File:  osd_file_manip.h
 * constants,structs and functions for reading and writing
 * from object file and debugging utilites.
 * Presently functions are for User Mode only.
 **/

#ifndef _KERNEL_
#include <stdint.h>
#include <byteswap.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <sys/uio.h>
/** Macros for Converting datatype from host to network */

#ifndef _BIG_ENDIAN
	#define be16_to_cpu(src)   bswap_16(src)
	#define be32_to_cpu(src)   bswap_32(src)
	#define be64_to_cpu(src)   bswap_64(src)
	#define cpu_to_be16(src)   bswap_16(src)
	#define cpu_to_be32(src)   bswap_32(src)
	#define cpu_to_be64(src)   bswap_64(src)
#else
	#define be16_to_cpu(src)   (src)
	#define be32_to_cpu(src)   (src)
	#define be64_to_cpu(src)   (src)
	#define cpu_to_be16(src)   (src)
	#define cpu_to_be32(src)   (src)
	#define cpu_to_be64(src)   (src)
#endif /** ifndef  _BIG_ENDIAN */


#endif /* ifdef _KERNEL_ */ 


#define FILENAME_SIZE 100
#define FILE_RW_BUFF_SIZE 1024
#define DIRECTORY_MODE 777

extern char k_buff[2048];
extern int DEBUG;

/**DEBUG Message Levels *****/
#define ERROR_MSG 0
#define UPPER_FUNC 1
# define LOWER_FUNC 2

/* Macros for printing debugging messages according to debug level */
#define print(args...) {  if(dlevel<=DEBUG){ \
sprintf(k_buff, args); \
printf("%s\n",k_buff);        }          \
}

#define print_msg(level,args...) {int i;   if(level<=DEBUG){ \
sprintf(k_buff, args);  for(i=0;i<level;i++) printf("	");  \
printf("%s:%s:%d:%s\n",                                     \
 __FILE__, __FUNCTION__, __LINE__, k_buff);        }          \
}
/* Macros for allocation/deallocation of buffere */
#define allocate(size) calloc(size,1)
#define free(buf)  free(buf);


/* Structure  to define data buffer */
typedef struct
{
	struct iovec *iov;
	int iov_count;
}OSD_DATA_IOV;

/** Structure for manipulation of collection list operations **/
typedef struct 
{
 int final;
 uint64_t coll_part_id;
 uint64_t coll_id;
 int object_type;
 uint64_t continuation_id;
 //uint8_t *data_buf;
 OSD_DATA_IOV *data_buf;
 uint32_t initial_data_offset;
 uint32_t data_buf_size;
 uint64_t object_count;/**From Quota attr page **/
}OSD_LIST_COMMAND_REPLY;


# define PRINT_OSD_LIST_COMMAND_REPLY_PTR(list_reply_ptr) \
print("Printing OSD_LIST_COMMAND_REPLY\n"); \
print("coll_part_id =0x%llx  ",list_reply_ptr->coll_part_id); \
print("coll_id =0x%llx  ",list_reply_ptr->coll_id); \
print("object_type =0x%x  ",list_reply_ptr->object_type); \
print("continuation_id =0x%llx  ",list_reply_ptr->continuation_id); \
print("data_buf_size =0x%lx  ",list_reply_ptr->data_buf_size); \
print("object_count =0x%llx \n",list_reply_ptr->object_count); \
print("initial_data_offset =0x%lx \n",list_reply_ptr->initial_data_offset);

# define TO_IOV_BUF 0
# define FROM_IOV_BUF 1

/**
 *
 *    Function:     iov_file_copy
 *
 *    Description:  Functions to transfer of data between OSD_DATA_IOV buffer and *		    user buffer
 *
 *    Arguments:    @data_buf - Program data buffer in OSD_DATA_IOV
 *                  @buf - data buffer array
 *                  @data_buf_offset - offset in data buffer
 *                  @length - num of bytes to read
 *                             legal value: length <= 64K (65536)
 * 		    @copy_direction - TO OSD_DATA_IOV(0) or From 	
 *					OSD_DATA_IOV(1)	
 *    Return codes: total num of bytes copied
 *
 **/

uint32_t iov_bufcpy(OSD_DATA_IOV *data_buf,uint8_t *buf,
			uint32_t data_buf_offset,uint32_t length,
			int copy_direction);



/** Object File manipulation operations **/

/**
 *
 *    Function:     get_attr_filename
 *
 *    Description:  Get Attribute file for object of give type
 *
 *    Arguments:    @part_id - Part_id for object
 *                  @object_id - Object_id for object
 *                  @page_no - Page_no 
 *                  @filename - File name returned for the object
 *  		    @object_type- Type of object 
 *					It can be
 *					UOBJ_OBJECT_TYPE 0x80 
 *					COLL_OBJECT_TYPE 0x40
 *					PART_OBJECT_TYPE 0x2 
 *					ROOT_OBJECT_TYPE 0x1
 *				
 *    Return codes: 0 if file exists,1 otherwise 
 *
 **/

int get_attr_filename(uint64_t part_id,uint64_t object_id, uint32_t page_no,
			char * filename,int object_type);




/**
 *
 *    Function:     get_data_filename
 *
 *    Description:  Get Data file for object of give type
 *
 *    Arguments:    @part_id - Part_id for object
 *                  @object_id - Object_id for object
 *                  @page_no - Page_no 
 *                  @filename - File name returned for the object
 * 		    @object_type- Type of object 
 *					It can be
 *					UOBJ_OBJECT_TYPE 0x80 
 *					COLL_OBJECT_TYPE 0x40
 *					PART_OBJECT_TYPE 0x2 
 *					ROOT_OBJECT_TYPE 0x1
 *				
 *    Return codes: 0 if file exists,1 otherwise 
 *
 **/

int get_data_filename(uint64_t part_id,uint64_t object_id, uint32_t page_no,
			char * filename,int object_type);


/**
 *
 *    Functions:    file_read/file_write *
 *    Description:  Read/Write from  object of give type
 *
 *    Arguments:    @filename - object file name
 *                  @buffer - data buffer
 *                  @size - read/write size
 *                  @offset - offset in file
 *    Return codes: no of bytes read/written 
 *
 **/

int file_write(char *filename,char *buffer,unsigned size,long offset);
int file_read(char *filename,char *buffer,unsigned size,long offset);




/**
 *
 *    Function:     object_exists 
 *    Description:  Check whether give object exists 
 *
 *    Arguments:    @part_id - partition id of object
 *                  @object_id - object id of object
 *                  @object_type - Type of object 
 *					It can be
 *					UOBJ_OBJECT_TYPE 0x80 
 *					COLL_OBJECT_TYPE 0x40
 * 					PART_OBJECT_TYPE 0x2 
 *					ROOT_OBJECT_TYPE 0x1
 *    Return codes: 1 if exist , 0 otherwise 
 *
 **/

int object_exists(uint64_t part_id,uint64_t object_id,int object_type);


/**
 *
 *    Functions:    Create/remove object data/attr file
 *    Description:  Create/Remove data/attr file of the given object
 *
 *    Arguments:    @part_id - partition id of object
 *                  @object_id - objct id of object
 *                  @object_type - Type of object 
 *					It can be
 *					UOBJ_OBJECT_TYPE 0x80 
 *					COLL_OBJECT_TYPE 0x40
 * 					PART_OBJECT_TYPE 0x2 
					
 *    Return codes: 0 if success , 1 otherwise 
 *
 **/

int create_object_data_file(uint64_t part_id,uint64_t object_id,int object_type);
int remove_object_data_file(uint64_t part_id,uint64_t object_id,int object_type);
int remove_object_attr_file(uint64_t part_id,uint64_t object_id,int object_type);





/**
 *
 *    Function:    perform_rw_operation
 *    Description: Perform Read/Write operation from object file to DATA_IOV	
 *		   This is upper level function invoked in OSD Operations which gets 
 *		   data file name of the object and writes data from OSD_DATA_IOV to * 		   the file				
 *
 *    Arguments:   @part_id - partition id of object
 *                 @object_id - object id of object
 *                 @data_iov - Program data iov buffer
 *		   @iovec_start_offset - data iov offset 
 *		   @obj_start_address - offset in object data
 *		   @length - read/write length
 *		   @rw - read (1)/write (0) direction	 	
 *   	 	
 *    Return codes: no of bytes read/written
 *
 **/

uint64_t perform_rw_operation(uint64_t part_id,uint64_t object_id,
				OSD_DATA_IOV *data_iov,uint64_t iovec_start_offset, uint64_t obj_start_address,uint64_t length,int rw);


/**
 *
 *    Functions:    add/remove_object_to_file
 *    Description:  Add/Remove object to/from collection object file 
 *			Called by  functions
 *			add_object_to_collection_list
 *			remove_object_from_collection_list		
 *
 *    Arguments:    @filename - Name of collection object data file
 *                  @add/rem_obj_id - add/remove object id of object
 *
 *    Return codes: 0 if success , 1 otherwise 
 *
 **/

int add_object_to_file(const char *filename,uint64_t add_obj_id);
int remove_object_from_file(const char *filename,uint64_t rem_obj_id);

/**
 *
 *    Functions:    get_random_object_id    
 *    Description:  Generate Random 64 bit no for object ID 
 *
 *    Arguments:    @object_type - Type of object
 *
 *    Return codes: 64 bit random object ID 
 *
 **/


uint64_t get_random_object_id(int object_type);

/**  Collection Object related operations **/

/**
 *
 *    Functions:   add/remove_object_to/from_collection_list 
 *    Description:  Add/Remove object to/from collection list
 *
 *    Arguments:    @coll_part_id - collection's partition id 
 *                  @coll_id - collection object id 
 *                  @add/rem_obj_id - Object id to add/remove 
 *		    @add/rem object_type - Type of object to add/remove
 *					It can be
 *					UOBJ_OBJECT_TYPE 0x80 
 *					COLL_OBJECT_TYPE 0x40
 * 					PART_OBJECT_TYPE 0x2 
 *			
 *    Return codes: 0 if success , 1 otherwise 
 *
 **/

int add_object_to_collection_list(uint64_t coll_part_id,uint64_t coll_id,
				uint64_t add_obj_id,int rem_object_type);
int remove_object_from_collection_list(uint64_t coll_part_id,uint64_t coll_id,
				uint64_t rem_obj_id,int rem_object_type);


/**
 *
 *    Functions:   list_objects_from_collection_list  
 *    Description:  Lists objects of given collection in data buffer
 *			OSD_LIST_COMMAND_REPLY gives 
 *			- Collection ID info (coll_part_id,coll_id,object type )
 *			- pointer to data_buf ,initial data offset,data buffer size
 *			Function returns
 *			final - set if all data list is copied to data buffer
 *			data_buf - data copied to data buf
 *			object_count - Total no of objects from list copied 	
 *
 *    Arguments:    @list_replay - Strcture for input/output of collection list  *				information
 *                  
 *    Return codes: 0 if success , 1 otherwise 
 *
 **/
int list_objects_from_collection_list(OSD_LIST_COMMAND_REPLY *list_replay);


/**
 *
 *    Functions:   print_objects_from_collection_list  
 *    Description:  Print objects from given collection data buffer to terminal
 *			It uses following fields from OSD_LIST_COMMAND_REPLY
 *			data_buf - data copied to data buf
 *			object_count - Total no of objects from list copied 	
 *			
 *
 *    Arguments:    @list_replay - Strcture for input/output of collection list   *				information
 *                  
 *    Return codes: 0 if success , 1 otherwise 
 *
 **/

int print_objects_from_collection_list(OSD_LIST_COMMAND_REPLY *list_replay);


/**
 *
 *    Functions:   is_uobject_member_of_collection
 *    Description: Checks whethr given object is member of collection  
 *
 *    Arguments:    @part_id - partition id of object
 *		    @obj_id - given member object id to check
 *		    @coll_obj_id - collection object id 	
 *
 *    Return codes: 0 if given object is member , 1 otherwise 
 *
 **/

int is_uobject_member_of_collection(uint64_t part_id,uint64_t obj_id,
				uint64_t coll_obj_id);




//int update_uobj_collections_page(uint64_t obj_id,uint64_t coll_obj_id);

/**
 *
 *    Function:     iov_file_copy
 *
 *    Description:  Copy collection list from collection file to data buffer
 *
 *    Arguments:    @fd - file descriptor for collection object data file
 *                  @data_iov - data buffer
 *                  @initia_offset - offset in data buffer
 *                  @read_len - num of bytes to read
 *                             legal value: length <= 64K (65536)
 * 
 *    Return codes: total num of bytes read
 *
 **/

uint32_t iov_file_copy(int fd,OSD_DATA_IOV *data_iov,uint32_t initial_offset,uint32_t read_len);


#endif /* ifdef OSD_FILE_MANIP_H */
