/**
 *                      ==================
 *                      IITK OSD Simulator
 *                      ==================
 *   An Implementation of Simulator for Object based storage device(OSD)
 *
 *   Copyright (C) 2007-2008 CSE IIT KANPUR
 *
 *   Contact: Department of Computer Science and Engineering
 *            Indian Institute of Technology,Kanpur(IIT Kanpur)
 *            {moona,riyaz}@cse.iitk.ac.in
 *
 *   This file is part of IITK OSD Simulator Project.
 *
 *   IITK OSD Simulator Project is free software; you can redistribute it
 *   and/or modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   IITK OSD Simulator Project is distributed in the hope that it will be
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the  GNU General
 *   Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

# include "osd_ops.h"
# include "osd_sense.h"
# include "osd_export.h"


/**
**********************************
 * File:  osd_ops.h
 * Implements functions for processing OSD operations
 **/

//Define debug level for this file as 1
static int dlevel=UPPER_FUNC;
extern int DEBUG;


/**
 *
 *    Function:    update_current_command_page
 *
 *    Description:  Update /Recreate Current command page according curernt OSD command
 * 
 *    Arguments:    @osd_req - OSD Req
 *		    @starting_byte_of_append - Offset for append command
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/
static inline int update_current_command_page(OSD_REQUEST *osd_req,u64 starting_byte_of_append)
{
	CURRENT_COMMAND_ATTRIBUTE_PAGE curr_cmd_attr_page;
	curr_cmd_attr_page.Object_Type=osd_req->object_type;
	curr_cmd_attr_page.Partition_ID=osd_req->part_id;
	curr_cmd_attr_page.CollectionUser_Object_ID=osd_req->object_id;
	curr_cmd_attr_page.Starting_Byte_Address_of_Append=starting_byte_of_append;
	memcpy(curr_cmd_attr_page.Response_Integirty_Chk_Value,osd_req->resp_int_chk_value,OSD_ATTR_NUM_CUR_CMD_RESPONSE_INT_CHECK_VAL_size);
	create_command_attribute_page(curr_cmd_attr_page);
	return 0;
}

//----------------------------------------------------------------------
//**************    DATA IOVEC MANIPULATION FUNCS  *********************
//----------------------------------------------------------------------
int osd_data_iov_alloc(OSD_DATA_IOV *data_buf,u32 size,u16 req_iov_size)
{

	int iov_count,i,iov_size,j;
	iov_count=(size+(req_iov_size-1))/req_iov_size;
	
	data_buf->iov=(struct iovec *)allocate(iov_count*sizeof(struct iovec));
	
	if(data_buf->iov == NULL)
		return 1;
	data_buf->iov_count=iov_count;
	
	for(i=0;i<iov_count;i++)
	{
		iov_size=req_iov_size;
		data_buf->iov[i].iov_base=(void *)allocate(iov_size);	
		if(data_buf->iov[i].iov_base == NULL)
		{
			for(j=0;j<i;j++)
			{
				free(data_buf->iov[j].iov_base);
			}
			return 1;
		}
		
		data_buf->iov[i].iov_len=iov_size;		
		size-=iov_size;	
	}
	return 0;
}
//----------------------------------------------------------------------
int osd_data_iov_free(OSD_DATA_IOV *data_buf)
{
	int i;

	if(data_buf->iov ==NULL)
	return 0;
 	for(i=0;i<data_buf->iov_count;i++)
 	{	free(data_buf->iov[i].iov_base);
	}
	free(data_buf->iov);
	data_buf->iov_count=0;
	return 0;
}
//----------------------------------------------------------------------
u32 iov_bufcpy(OSD_DATA_IOV *data_buf,u8 *buf,u32 data_buf_offset,u32 length,int copy_direction)
{
	int i;
	u32 curr=0,iov_remain=0;
	u32 iov_offset,copy_len=0;
	for(i=0;i< data_buf->iov_count;i++)
	{
		if(curr + data_buf->iov[i].iov_len > data_buf_offset )
		{
			break;
		} 
		curr+=data_buf->iov[i].iov_len;
	}
	
	if(i==data_buf->iov_count)
	{
		print_msg(ERROR_MSG,"Serious internal buffer allocation error");
		return 0;
	}
	
	iov_offset=data_buf_offset-curr;
	print_msg(dlevel+1,"iov buf found =0x%x with offset =0x%lx length =0x%lx\n",i,iov_offset,length);	
	while(copy_len < length && i < data_buf->iov_count)
	{
		iov_remain=data_buf->iov[i].iov_len-iov_offset;
		if(copy_len+iov_remain >= length)
		{
			iov_remain=length-copy_len;
			if(copy_direction==TO_IOV_BUF)
			memcpy((u8 *)data_buf->iov[i].iov_base+iov_offset,(u8 *)buf+copy_len,iov_remain);
			else
			memcpy((u8 *)buf+copy_len,(u8 *)data_buf->iov[i].iov_base+iov_offset,iov_remain);
			copy_len +=iov_remain;
			break;
		}
		else
		{
		if(copy_direction==TO_IOV_BUF)
		memcpy((u8 *)data_buf->iov[i].iov_base + iov_offset,(u8 *)buf+copy_len,data_buf->iov[i].iov_len);
		else
		memcpy((u8 *)buf+copy_len,(u8 *)data_buf->iov[i].iov_base + iov_offset,data_buf->iov[i].iov_len);

		copy_len +=data_buf->iov[i].iov_len;
		}	
		
		i++;
		iov_offset=0;
	}
	print_msg(dlevel+1,"Data copied len=0x%lx",copy_len);
	return copy_len;
}
//----------------------------------------------------------------------
void *get_osd_data_iov_offset(OSD_DATA_IOV *data_buf,u32 offset)
{
	int i;
	void *ptr;
	u32 curr=0,off;

	for(i=0;i< data_buf->iov_count;i++)
	{


		if(curr + data_buf->iov[i].iov_len > offset)
		{
			
			off=offset-curr;
			ptr = (u8 *)data_buf->iov[i].iov_base+off;
			return ptr;
		}
		curr+=data_buf->iov[i].iov_len; 
	}

	return NULL;
}
//----------------------------------------------------------------------
//**********  Other OSD  MANIPULAATION FUNCS **************************
//----------------------------------------------------------------------
int object_existance_test(OSD_REQUEST *osd_req,u64 part_id,u64 obj_id,int object_type)
{

       if(!object_exists(part_id,0,PART_OBJECT_TYPE))
        {
                print_msg(ERROR_MSG,"Partition with part_id=0x%llx   does not exists or its < 0x10000",part_id);
                osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
                create_osd_error_id_sense_data_desc(osd_req);
                //offset req_object_id in CDB=16
                create_osd_key_specific_sense_data_desc(osd_req,16,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
                return 1;
        }
	if(obj_id)
	{
		if(!object_exists(part_id,obj_id,object_type))
		{
			if(object_type == COLL_OBJECT_TYPE)
			{
			print_msg(ERROR_MSG,"Collection Object with part_id=0x%llx object_id =0x%llx does not exists",part_id,obj_id);
			}
			else
			{
			print_msg(ERROR_MSG,"Collection Object with part_id=0x%llx object_id =0x%llx does not exists",part_id,obj_id);
			}
			
			osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
			create_osd_error_id_sense_data_desc(osd_req);
			//offset rem_collection_object_id in CDB=24
			create_osd_key_specific_sense_data_desc(osd_req,24,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
			
			return 1; 
		
		}
	}

return 0;

} 
//----------------------------------------------------------------------
//**********  OSD REQUEST MANIPULAATION FUNCS **************************
//----------------------------------------------------------------------
OSD_REQUEST *allocate_osd_request()
{
	OSD_REQUEST *osd_req;
	
	osd_req=(OSD_REQUEST *)allocate(sizeof(OSD_REQUEST));
	if(osd_req==NULL)
	return NULL;

	osd_req->data_in.iov_count=0;
	osd_req->data_out.iov_count=0;
	osd_req->get_attr_list=NULL;
	osd_req->get_attr_list_count=0;
	osd_req->set_attr_list=NULL;
	osd_req->set_attr_list_count=0;
	osd_req->set_attr_list_data_buf_size=0;
	osd_req->get_attr_list_data_buf_size=0;
	osd_req->osd_lun=NULL;
	osd_req->data_in.iov=NULL;
	osd_req->data_out.iov=NULL;
	osd_req->data_iov_size=0;
	
 //	osd_req->osd_sense_response.status=0;
	osd_req->osd_sense_response.sense_data_size=0;
	memset(osd_req->completed_command_functions,4,0);
 	memset(osd_req->not_initiated_command_functions,4,0);
	osd_req->osd_sense_response.sense_buf=(u8 *)allocate(MAX_SENSE_BUF_SIZE);
	
	memset(osd_req->resp_int_chk_value,0,OSD_ATTR_NUM_CUR_CMD_RESPONSE_INT_CHECK_VAL_size);
	if(osd_req->osd_lun)
		free(osd_req->osd_lun);
	return osd_req;
}

int set_osd_req_lun(OSD_REQUEST *osd_req,LUN_STRUCT *lun_struct)
{
	osd_req->osd_lun=allocate(sizeof(OSD_LUN));
	if(get_osd_lun(lun_struct,osd_req->target_name,osd_req->lun_no,osd_req->osd_lun))
	{
		print_msg(ERROR_MSG,"Error in Target not found with name=%s Lun=%d ",osd_req->target_name,osd_req->lun_no);
		return 1;
	}
	else
	{
		print_msg(dlevel,"Target Name Found = %s",osd_req->target_name);
	
	}

return 0;	

}
//----------------------------------------------------------------------

int free_osd_request(OSD_REQUEST * osd_req)
{

	osd_data_iov_free(&(osd_req->data_in));
	osd_data_iov_free(&(osd_req->data_out));
	osd_req->get_attr_list=NULL;
	osd_req->get_attr_list_count=0;
	osd_req->set_attr_list=NULL;
	osd_req->set_attr_list_count=0;
	osd_req->osd_sense_response.sense_data_size=0;
	free(osd_req->osd_sense_response.sense_buf);
	free(osd_req);
	return 0;
}

//----------------------------------------------------------------------
//**********  OSD GET-SET ATTR MANIPULATION FUNCS **********************
//----------------------------------------------------------------------

int extract_get_attribute_list(OSD_REQUEST *osd_req)
{
    int i;
    u32 total_attrs;
    u32 no;

    u32 get_list_length,get_list_din_offset;
    OSD_ATTR_LIST *get_attr_list;
    OSD_DATA_IOV *data_buf;
    OSD_CDB_ATTR_LIST_HDR *osd_cdb_attr_list_hdr;
    osd_cdb_attr_list_hdr=(OSD_CDB_ATTR_LIST_HDR *)&(osd_req->osd_cdb_struct);

    get_list_length=osd_cdb_attr_list_hdr->get_list_length;
    get_list_din_offset=osd_cdb_attr_list_hdr->get_list_dout_offset;
   
    //Since first 4 bytes in attr list format contains list-type list lenght
    //As of now we simply ignore this
    print_msg(dlevel,"get_list_din_offset =0x%lx (+4)  get_list_length=0x%lx (-4) ",get_list_din_offset,get_list_length);
    get_list_din_offset +=4;
    get_list_length-=4;

    total_attrs= get_list_length/8;

    print_msg(dlevel,"total_attrs =0x%lx sizeof(OSD_ATTR_LIST)=0x%x  MAX_GET_LIST_HDR_SIZE=0x%x ",total_attrs,sizeof(OSD_ATTR_LIST),MAX_GET_LIST_HDR_SIZE);
    
    if(total_attrs * sizeof(OSD_ATTR_LIST) > MAX_GET_LIST_HDR_SIZE)
    {
    	print_msg(ERROR_MSG," No of attributes are too large to fit in Get Attribute hdr list");
    	return 1;

    }
    osd_req->get_attr_list=(OSD_ATTR_LIST *)allocate(MAX_GET_LIST_HDR_SIZE);

    if(osd_req->get_attr_list ==NULL)
    {
    	print_msg(ERROR_MSG,"Error allocating GET_LIST_HDR Buffer");
    
	return 1;
    }
    
    get_attr_list=osd_req->get_attr_list;
    data_buf=&(osd_req->data_in);

	/** For each entry in get list request,
	saves retrieved/get attribute entry(page_no,attr_no)
	in get_attr_list buffer of osd_req 
	Note (attr_len,attr_values) will be set in function process_get_attribute_list function 
	*/ 
    
    for(i=0;i<total_attrs;i++)
    {
   	//Get Page no from OSD_IOV_BUF data_buf to no 
	if(iov_bufcpy(data_buf,(u8 *) &no,get_list_din_offset,4,FROM_IOV_BUF)!=4)
	{
		print_msg(ERROR_MSG,"Problem getting offset for page_no");
		return 1; 
	}

	//Quota error can be  here

	get_attr_list[i].attr_din_offset=get_list_din_offset;
	get_attr_list[i].page_no=be32_to_cpu(no);
	get_list_din_offset+=4;

	//Get Attr no from OSD_IOV_BUF data_buf to no
	if(iov_bufcpy(data_buf,(u8 *)&no,get_list_din_offset,4,FROM_IOV_BUF)!=4)
	{
		print_msg(ERROR_MSG,"Problem getting offset for attr_no");
		return 1; 
	}
	//memcpy(&(get_attr_list[i].attr_no),ptr,4);
	get_attr_list[i].attr_no=be32_to_cpu(no);
	get_list_din_offset+=4;

	{
		u32 page_no=get_attr_list[i].page_no,attr_no=get_attr_list[i].attr_no;
		print_msg(dlevel,"Get Attribute entry page_no=0x%lx  attr_no=0x%lx ",page_no,attr_no);
	}
    }

    osd_req->get_attr_list_count=total_attrs;

    return 0;
}
//----------------------------------------------------------------------

int process_get_attribute_list(OSD_REQUEST *osd_req)
{

	int i,len;
	OSD_ATTR_INFO osd_attr_info;
	u32 dout_buf_size;
	u32 get_list_data_buf_offset=0;
	u32 get_alloc_length;//need to set
	u32 retrieved_attr_dout_offset;
	OSD_CDB_ATTR_LIST_HDR *osd_cdb_list_hdr;
	OSD_ATTR_LIST *get_attr_list=osd_req->get_attr_list;
	u8 attr_value[MAX_ATTR_SIZE];
	


	osd_cdb_list_hdr=(OSD_CDB_ATTR_LIST_HDR *)&(osd_req->osd_cdb_struct);
	get_alloc_length=osd_cdb_list_hdr->get_allocation_length;	
	retrieved_attr_dout_offset=osd_cdb_list_hdr->retrieved_attr_din_offset;
	
	osd_req->get_attr_list_data_buf=(u8 *)allocate(MAX_GET_LIST_DATA_SIZE);
	
	//set get list header
	//set list type
	osd_req->get_attr_list_data_buf[0]=(u8)0x9;	
	get_list_data_buf_offset+=4;
	
	if(osd_req->get_attr_list_data_buf==NULL)
	{
		print_msg(ERROR_MSG,"Error allocating SET_LIST DATA Buffer");
		return 1;
	}
	print_msg(dlevel,"get attr_list_count =0x%lx",osd_req->get_attr_list_count);
	
	/** For each entry in get list request,
	saves retrieved attribute entry(page_no,attr_no,attr_len,attr_value)
	in get_attr_list buffer of osd_req 
	Note (page_no,attr_no) are set in function extract_get_attribute_list function 
	*/ 
	for(i=0;i< osd_req->get_attr_list_count;i++)
	{
		osd_attr_info = get_osd_attr_struct(osd_req->part_id,osd_req->object_id,get_attr_list[i].page_no,get_attr_list[i].attr_no,osd_req->object_type);
		
		{
		u32 page_no=get_attr_list[i].page_no,  attr_no=get_attr_list[i].attr_no;
		print_msg(ERROR_MSG,"Attr details: part_id=0x%llx obj_id=0x%llx  page_no=0x%lx attr_no=0x%lx object_type=0x%x",osd_req->part_id,osd_req->object_id,page_no,attr_no,osd_req->object_type);
		}
		print_msg(dlevel,"Got osd attr_struct as  attr_no=0x%lx attr_size=0x%x attr_offset=0x%x",osd_attr_info.attr_no,osd_attr_info.attr_size,osd_attr_info.attr_offset);
		
	
		if (osd_attr_info.attr_offset ==OSD_ATTR_WRONG_OFFSET) //This is error either page/Attr no invalid
		{	//this is error
			//offset of attribute in din buffer
			u32 offset;
    			print_msg(ERROR_MSG,"Error in getting offset for given attribute ");
			print_msg(ERROR_MSG,"Attr details: part_id=0x%llx obj_id=0x%llx  page_no=0x%lx attr_no=0x%lx object_typ=0x%x",osd_req->part_id,osd_req->object_id,get_attr_list[i].page_no,get_attr_list[i].attr_no,osd_req->object_type);
			
			osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_PARAMETER_LIST_ASC,SCSI_INVALID_FIELD_IN_PARAMETER_LIST_ASCQ,osd_req->osd_response_header.data_len);
			create_osd_error_id_sense_data_desc(osd_req);
			//offset of attribute in din buffer
			offset=(u32)get_attr_list[i].attr_din_offset;
			create_osd_key_specific_sense_data_desc(osd_req,(u16)offset,OSD_SENSE_KEY_SPF_DATA_FIELD_FLAG);
	
			getattr_attribute_pages_effects(osd_req->part_id,osd_req->object_id,osd_req->object_type);
			return 1;
		}
	
		get_attr_list[i].attr_len=osd_attr_info.attr_size;
		get_attr_list[i].attr_val_offset=get_list_data_buf_offset;
		
		len=get_list_data_buf_offset +osd_attr_info.attr_size;
		if(len > get_alloc_length || len > MAX_GET_LIST_DATA_SIZE)
		{
			goto finish_get_attr_process;
			//this is not error according to OSD specification document
		}
		
		{
			u32 temp32;
			u16 temp16;
			
			temp32=cpu_to_be32(get_attr_list[i].page_no);
			memcpy((u8 *)osd_req->get_attr_list_data_buf+ get_list_data_buf_offset,&temp32,4);
			get_list_data_buf_offset+=4;
			
			temp32=cpu_to_be32(get_attr_list[i].attr_no);
			memcpy((u8 *)osd_req->get_attr_list_data_buf +get_list_data_buf_offset,&temp32,4);
			get_list_data_buf_offset+=4;
			
			temp16=cpu_to_be16(osd_attr_info.attr_size);
			memcpy((u8 *)osd_req->get_attr_list_data_buf +get_list_data_buf_offset,&temp16,2);
			get_list_data_buf_offset+=2;
			
		}
		
		get_page_attr(osd_req->part_id,osd_req->object_id,get_attr_list[i].page_no,get_attr_list[i].attr_no,attr_value,osd_req->object_type);
		
		//Convert from attr value Host representation to network
		if(osd_attr_info.attr_size== 8)
		{
			u64 tmp;tmp=cpu_to_be64(*((u64 *)attr_value));
			memcpy((u8 *)osd_req->get_attr_list_data_buf +get_list_data_buf_offset,&tmp,osd_attr_info.attr_size);
		}else if(osd_attr_info.attr_size== 4)
		{
			u32 tmp;tmp=cpu_to_be32(*((u32 *)attr_value));
			memcpy((u8 *)osd_req->get_attr_list_data_buf +get_list_data_buf_offset,&tmp,osd_attr_info.attr_size);
			
		}else if(osd_attr_info.attr_size== 2)
		{
			u16 tmp;tmp=cpu_to_be16(*((u16 *)attr_value));
			memcpy((u8 *)osd_req->get_attr_list_data_buf +get_list_data_buf_offset,&tmp,osd_attr_info.attr_size);
		}
		else if(osd_attr_info.attr_size== 6)
		{
			u48 tmp; tmp=cpu_to_be48(*((u48 *)attr_value));
			memcpy((u8 *)osd_req->get_attr_list_data_buf +get_list_data_buf_offset,&tmp,osd_attr_info.attr_size);
		}
		else
			memcpy((u8 *)osd_req->get_attr_list_data_buf +get_list_data_buf_offset,attr_value,osd_attr_info.attr_size);
		
		get_list_data_buf_offset +=osd_attr_info.attr_size;
		
	}

	*(u16 *)(osd_req->get_attr_list_data_buf +2)=cpu_to_be16(get_list_data_buf_offset-4);	
		
finish_get_attr_process:		
	len=iov_bufcpy(&(osd_req->data_out),osd_req->get_attr_list_data_buf, retrieved_attr_dout_offset,get_list_data_buf_offset,TO_IOV_BUF);
	if(len != get_list_data_buf_offset)
	{	//Its internal error
		print_msg(ERROR_MSG,"Error in copying from get_attr_data_but to data_out iovec ");
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_PARAMETER_LIST_ASC,SCSI_INVALID_FIELD_IN_PARAMETER_LIST_ASCQ,osd_req->osd_response_header.data_len);
		create_osd_error_id_sense_data_desc(osd_req);
		//offset of attribute in din buffer
		create_osd_key_specific_sense_data_desc(osd_req,(u16)get_list_data_buf_offset,OSD_SENSE_KEY_SPF_DATA_FIELD_FLAG);
		return 1;
	}

	getattr_attribute_pages_effects(osd_req->part_id,osd_req->object_id,osd_req->object_type);
		
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_GET_ATTR_BYTE,COMMAND_FUNCTIONS_GET_ATTR);
	
	//its time to update  data_out_buffer data_len 
	//255
	//This will be always at the end of data
	dout_buf_size= osd_cdb_list_hdr->retrieved_attr_din_offset+get_list_data_buf_offset; 
	dout_buf_size+=255;
	dout_buf_size &= ~(255);
	
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,dout_buf_size);

	return 0;

}
//------------------------------------------------------------------------
int extract_set_attribute_list(OSD_REQUEST *osd_req)
{
	//void *ptr;
	u8 *set_attr_list_data_buf;
	OSD_ATTR_LIST *set_attr_list;
	OSD_CDB_ATTR_LIST_HDR *osd_cdb_attr_list_hdr;
    	u32 set_list_length,set_list_din_offset;
	int entry_count=0;
	u32 set_list_data_buf_offset=0;
	OSD_DATA_IOV *data_buf;
	u32 no;u16 len;
	u64 val64;u32 val32;u16 val16;
	osd_req->set_attr_list_data_buf=(u8 *)allocate(MAX_SET_LIST_DATA_SIZE);
	
	if(osd_req->set_attr_list_data_buf==NULL)
	{
		print_msg(ERROR_MSG,"Error allocating SET_LIST DATA Buffer");
	
		return 1;
	}
	
	set_attr_list_data_buf=osd_req->set_attr_list_data_buf;
	
	osd_req->set_attr_list=(OSD_ATTR_LIST *)allocate(MAX_SET_LIST_HDR_SIZE);
	
	if(osd_req->set_attr_list ==NULL)
	{
		print_msg(ERROR_MSG,"Error allocating GET_LIST_HDR Buffer");
	
		return 1;
	}
	set_attr_list=osd_req->set_attr_list;
	
	
	
	osd_cdb_attr_list_hdr=(OSD_CDB_ATTR_LIST_HDR *)&(osd_req->osd_cdb_struct);
    
    	set_list_length=osd_cdb_attr_list_hdr->set_list_length;
    	set_list_din_offset=osd_cdb_attr_list_hdr->set_list_dout_offset;
    
	data_buf=&(osd_req->data_in);
	
	print_msg(dlevel,"set_list_din_offset =0x%lx (+4)  set_list_length=0x%lx (-4) ",set_list_din_offset,set_list_length);
    	set_list_din_offset +=4;
    	//set_list_length-=4;

	/** For each entry in set list request,
	saves set attribute entry(page_no,attr_no,attr_len,attr_value)
	in set_attr_list buffer of osd_req 
	*/ 
    	while(set_list_din_offset<set_list_length )
	{
		
		if((set_list_din_offset+ 10) > set_list_length ) //page_no size 4 attr_no size 4  attr_len 2
		{
			print_msg(ERROR_MSG,"Error in set list buffer sent from Client ");
			
			osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_PARAMETER_LIST_ASC,SCSI_INVALID_FIELD_IN_PARAMETER_LIST_ASCQ,osd_req->osd_response_header.data_len);
			create_osd_error_id_sense_data_desc(osd_req);
	
			//offset req_object_id in CDB=24
			create_osd_key_specific_sense_data_desc(osd_req,(u16)set_list_din_offset,OSD_SENSE_KEY_SPF_DATA_FIELD_FLAG);
			return 1;
		}

		if(iov_bufcpy(data_buf,(u8 *) &no,set_list_din_offset,4,FROM_IOV_BUF)!=4)
		{
				print_msg(ERROR_MSG,"Problem getting offset for page_no");
			return 1; 
		}

		//memcpy(&(set_attr_list[entry_count].page_no),ptr,4);
		set_attr_list[entry_count].page_no=be32_to_cpu(no);
		set_list_din_offset+=4;
	

		if(iov_bufcpy(data_buf,(u8 *) &no,set_list_din_offset,4,FROM_IOV_BUF)!=4)
		{
			print_msg(ERROR_MSG,"Problem getting offset for attr_no");
			return 1; 
		}

		set_attr_list[entry_count].attr_din_offset=set_list_din_offset;
		set_attr_list[entry_count].attr_no=be32_to_cpu(no);
		set_list_din_offset+=4;


		if(iov_bufcpy(data_buf,(u8 *) &len,set_list_din_offset,2,FROM_IOV_BUF)!=2)
		{
			print_msg(ERROR_MSG,"Problem getting offset for attr_len");
			return 1; 
		}
			
		set_attr_list[entry_count].attr_len=be16_to_cpu(len);
		set_list_din_offset+=2;
		print_msg(dlevel,"Set Attribute entry page_no=0x%lx  attr_no=0x%lx attr_len =0x%x ",set_attr_list[entry_count].page_no,set_attr_list[entry_count].attr_no,set_attr_list[entry_count].attr_len);
		
		if(set_list_data_buf_offset+set_attr_list[entry_count].attr_len >MAX_SET_LIST_DATA_SIZE)
		{
			print_msg(ERROR_MSG,"Set List Buffer too small to accept this data ");
			osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_PARAMETER_LIST_ASC,SCSI_INVALID_FIELD_IN_PARAMETER_LIST_ASCQ,osd_req->osd_response_header.data_len);
			create_osd_error_id_sense_data_desc(osd_req);
	
			//offset req_object_id in CDB=24
			create_osd_key_specific_sense_data_desc(osd_req,(u16)set_list_din_offset,OSD_SENSE_KEY_SPF_DATA_FIELD_FLAG);
			return 1;
			
		} 
		
		if(set_list_din_offset+set_attr_list[entry_count].attr_len >set_list_length)
		{
			print_msg(ERROR_MSG,"Set List length too small to get attribute  ");
			osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_PARAMETER_LIST_ASC,SCSI_INVALID_FIELD_IN_PARAMETER_LIST_ASCQ,osd_req->osd_response_header.data_len);
			create_osd_error_id_sense_data_desc(osd_req);
	
			//offset req_object_id in CDB=24
			create_osd_key_specific_sense_data_desc(osd_req,(u16)set_list_din_offset,OSD_SENSE_KEY_SPF_DATA_FIELD_FLAG);
		
		 return 1;
		 //
		}
		
		print_msg(dlevel,"Set list_data_buf offset= 0x%lx  set_list_din_offset= 0x%lx", set_list_data_buf_offset,set_list_din_offset);
			
		print_msg(dlevel,"Call to iov_bufcpy to get set list attr_value");
	
		//Convert attribute value from network represenation to host 
		if(set_attr_list[entry_count].attr_len==2)
		{
			if(iov_bufcpy(data_buf,(u8 *)&val16,set_list_din_offset,set_attr_list[entry_count].attr_len,FROM_IOV_BUF)!=set_attr_list[entry_count].attr_len)
			{
			print_msg(ERROR_MSG,"Problem getting offset for attr_value");
			return 1; 
			}
			val16=be16_to_cpu(val16);
			memcpy((u8 *) set_attr_list_data_buf+set_list_data_buf_offset,&val16,2);
		
			
		}
		else if(set_attr_list[entry_count].attr_len==4)
		{
			if(iov_bufcpy(data_buf,(u8 *)&val32,set_list_din_offset,set_attr_list[entry_count].attr_len,FROM_IOV_BUF)!=set_attr_list[entry_count].attr_len)
			{
			print_msg(ERROR_MSG,"Problem getting offset for attr_value");
			return 1; 
			}
			val32=be32_to_cpu(val32);
			memcpy((u8 *) set_attr_list_data_buf+set_list_data_buf_offset,&val32,4);
			
		}else if(set_attr_list[entry_count].attr_len==6)
		{
			u48 val48;	
			if(iov_bufcpy(data_buf,(u8 *)&val48,set_list_din_offset,set_attr_list[entry_count].attr_len,FROM_IOV_BUF)!=set_attr_list[entry_count].attr_len)
			{
			print_msg(ERROR_MSG,"Problem getting offset for attr_value");
			return 1; 
			}
			val48=be48_to_cpu(val48);
			memcpy((u8 *) set_attr_list_data_buf+set_list_data_buf_offset,&val48,6);

		}
		else if(set_attr_list[entry_count].attr_len==8)
		{
			if(iov_bufcpy(data_buf,(u8 *)&val64,set_list_din_offset,set_attr_list[entry_count].attr_len,FROM_IOV_BUF)!=set_attr_list[entry_count].attr_len)
			{
			print_msg(ERROR_MSG,"Problem getting offset for attr_value");
			return 1; 
			}
			val64=be64_to_cpu(val64);
			memcpy((u8 *) set_attr_list_data_buf+set_list_data_buf_offset,&val64,8);
		}
		
	
		set_attr_list[entry_count].attr_val_offset=set_list_data_buf_offset;
		set_list_data_buf_offset+=set_attr_list[entry_count].attr_len;
		//set_attr_list_data_buf+=set_attr_list[entry_count].attr_len;
		set_list_din_offset+=set_attr_list[entry_count].attr_len;
		entry_count++;
	}
	
	
	osd_req->set_attr_list_data_buf_size=set_list_data_buf_offset;
	osd_req->set_attr_list_count=entry_count;
	print_msg(dlevel,"Set Attr list count= 0x%lx ",osd_req->set_attr_list_count);
	
	return 0;
	

}
//------------------------------------------------------------------------
int process_set_attribute_list(OSD_REQUEST *osd_req)
{

	int i;
	OSD_ATTR_INFO osd_attr_info;
	OSD_ATTR_LIST *set_attr_list=osd_req->set_attr_list;
		
	if(osd_req->set_attr_list_data_buf==NULL)
	{
		//this error is never expected to happen
		print_msg(ERROR_MSG,"Error no SET_LIST DATA Buffer ");
		return 1;
	}
	
	/** For each set attribute entry from set_attr_list buffer of osd_req process set attribute */
	for(i=0;i< osd_req->set_attr_list_count;i++)
	{
	
		osd_attr_info = get_osd_attr_struct(osd_req->part_id,osd_req->object_id,set_attr_list[i].page_no,set_attr_list[i].attr_no,osd_req->object_type);
		
		//Note that osd_attr_info.attr_size=0 for vendor specific page attribute when set first time 
		if (osd_attr_info.attr_offset ==OSD_ATTR_WRONG_OFFSET ||(osd_attr_info.attr_size &&  (osd_attr_info.attr_size != set_attr_list[i].attr_len)) || !(osd_attr_info.attr_set))
		{	//This is error 
			//offset of attribute in din buffer
			//u32 offset32;
			print_msg(ERROR_MSG,"Error in setting offset for given attribute ");
	
			osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_PARAMETER_LIST_ASC,SCSI_INVALID_FIELD_IN_PARAMETER_LIST_ASCQ,osd_req->osd_response_header.data_len);
			create_osd_error_id_sense_data_desc(osd_req);
			//offset of attribute in din buffer
			
			if(osd_attr_info.attr_offset == OSD_ATTR_WRONG_OFFSET)
			{
				create_osd_key_specific_sense_data_desc(osd_req,(u16)(set_attr_list[i].attr_din_offset),OSD_SENSE_KEY_SPF_DATA_FIELD_FLAG);
			}
			else
			{
				create_osd_key_specific_sense_data_desc(osd_req,(u16)(set_attr_list[i].attr_din_offset+8),OSD_SENSE_KEY_SPF_DATA_FIELD_FLAG);
			}
			setattr_attribute_pages_effects(osd_req->part_id,osd_req->object_id,osd_req->object_type);
			return 1;
		}
	
		set_page_attr(osd_req->part_id,osd_req->object_id,set_attr_list[i].page_no,set_attr_list[i].attr_no,(u8 *)osd_req->set_attr_list_data_buf +set_attr_list[i].attr_val_offset,osd_req->object_type);
		
	}

	setattr_attribute_pages_effects(osd_req->part_id,osd_req->object_id,osd_req->object_type);
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_SET_ATTR_BYTE,COMMAND_FUNCTIONS_SET_ATTR);

	return 0;


}
//----------------------------------------------------------------------
u16 cmnd_opcode(OSD_REQUEST *osd_req)
{
	OSD_CDB_BASIC_HDR *osd_cdb_basic_hdr;
	osd_cdb_basic_hdr=(OSD_CDB_BASIC_HDR *)&(osd_req->osd_cdb_struct);
	return  osd_cdb_basic_hdr->service_action;
}

//----------------------------------------------------------------------
//**********  OSD COMMAND FUNCS **********************
//----------------------------------------------------------------------

int osd_format(OSD_REQUEST *osd_req)
{

	OSD_CDB_FORMAT_COMMAND_HDR *osd_cdb_command_format;
	int ret;
	osd_req->object_type=ROOT_OBJECT_TYPE;
	osd_req->part_id=0;
	osd_req->object_id=0;
	
	osd_cdb_command_format=(OSD_CDB_FORMAT_COMMAND_HDR *)&(osd_req->osd_cdb_struct);
	if(osd_cdb_command_format->formatted_capacity ==0)
		osd_cdb_command_format->formatted_capacity=osd_req->osd_lun->capacity;
	if(osd_req->osd_lun->capacity > osd_cdb_command_format->formatted_capacity)
	{
		print_msg(dlevel,"OSD: Presently, we dont have support for partial formatting,Formatting whole unit\n");
	
	} else if(osd_req->osd_lun->capacity < osd_cdb_command_format->formatted_capacity)
	{ //According to OSD-2 T10 document  section 6.11 , This is not error
		print_msg(dlevel,"OSD: Request format unit size=0x%llx Actual size=0x%llx ,truncating \n",osd_cdb_command_format->formatted_capacity,osd_req->osd_lun->capacity);
		osd_cdb_command_format->formatted_capacity=osd_req->osd_lun->capacity;
	}
	/** Implement OSD Format command as 
		- Deleting contents of OSD Lun Base directory 
		- Creating root and partittion 0 attribute pages
	*/	

# ifndef _KERNEL_
	char cmd[FILENAME_SIZE];	
	strcpy(cmd,"rm -fr ");
	strcat(cmd,osd_req->osd_lun->osd_lun_base_directory);
	 //Not Rm -fr never gives error except permissionn set problem
	print_msg(dlevel," Removing directory  %s \n",cmd);
	
	if(system(cmd)) 
	{
		print_msg(ERROR_MSG," Error executing format command \n");
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
		//FUTURE: this error type is not known
		create_osd_error_id_sense_data_desc(osd_req);
		return 1;
		
	
	}
	
	strcpy(cmd,"mkdir ");
	strcat(cmd,osd_req->osd_lun->osd_lun_base_directory);
	print_msg(dlevel," Creating directory  %s \n",cmd);
	if(system(cmd))	
	{
		print_msg(ERROR_MSG," Error executing format command \n");
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
		//FUTURE: this error type is not known
		create_osd_error_id_sense_data_desc(osd_req);
		return 1;
		
	}
	
	print_msg(dlevel," Data erased successfully\n");
#endif /** ifndef _KERNEL_ */
	
	/** Create Root and partition 0 attrbite pages */
	if(ret=osd_format_attribute_pages_effects(osd_req->osd_lun->osd_system_id,osd_req->osd_lun->capacity,osd_req->osd_lun->username))
	{
		print_msg(ERROR_MSG," Error Creating root attribute pages format command \n");
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_DATA_PROTECT,SCSI_QUOTA_ERROR_ASC,SCSI_QUOTA_ERROR_ASCQ,0);
		
		create_osd_error_id_sense_data_desc(osd_req);
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_ROOT_INFO,OSD_ATTR_NUM_ROOT_INFO_TOTAL_CAPACITY);		
		return 1;
	}
	print_msg(dlevel,"OSD root attr pages created successfully\n");
	
	/** To maintain list of all partition create root data file */
	if(create_object_data_file(0,0,ROOT_OBJECT_TYPE))
	{
		print_msg(ERROR_MSG," Error Creating Root object data file\n");
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_DATA_PROTECT,SCSI_QUOTA_ERROR_ASC,SCSI_QUOTA_ERROR_ASCQ,0);
		//FUTURE: this error type is not known
		create_osd_error_id_sense_data_desc(osd_req);
		return 1;
		
	
	}
	//Update command functions bit for command execution
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	print_msg(dlevel,"OSD Format  unit done successfully,updating command page\n");
	
	//Update Current_command page
	update_current_command_page(osd_req,0);	
	//Update command functions bit for set attr effects of command
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);
	//Build response header as SCSI GOOD
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,0);

	return 0;
}
//---------------------------------------------------------------------------------------------------
int osd_create_partition(OSD_REQUEST *osd_req)
{
	//OSD_RESPONSE_HEADER osd_response;
	OSD_CDB_CREATE_PARTITION_COMMAND_HDR *osd_cdb_command_create_part;
	u64 part_id;
	int ret;
	
	osd_cdb_command_create_part=(OSD_CDB_CREATE_PARTITION_COMMAND_HDR *)&(osd_req->osd_cdb_struct);
	print_msg(dlevel,"osd_cdb_command_create_part->req_part_id=0x%llx",osd_cdb_command_create_part->req_part_id);
	
	
	if(osd_cdb_command_create_part->req_part_id)
	{	//Check if part id already exists 
		if(object_exists(osd_cdb_command_create_part->req_part_id ,0,PART_OBJECT_TYPE) || osd_cdb_command_create_part->req_part_id < 0x10000)
		{
		print_msg(ERROR_MSG,"Part object with part_id=0x%llx already exists or its < 0x10000",osd_cdb_command_create_part->req_part_id);
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
		//offset req_partion_id in CDB=16
		create_osd_key_specific_sense_data_desc(osd_req,16,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
		return 1;
		
		}
		part_id=osd_cdb_command_create_part->req_part_id;
	}else
	{
	
		while(1)
		{	//Get random part id
			part_id=get_random_object_id(PART_OBJECT_TYPE);
			if(!object_exists(part_id,0,PART_OBJECT_TYPE))
			break;
			
		}
	
	}
	
	osd_req->object_type=PART_OBJECT_TYPE;
	osd_req->part_id=part_id;
	osd_req->object_id=0;

	print_msg(dlevel,"OSD partition id= 0x%llx\n",part_id);
	
	/** Create part id data file to save list of objects */
	if(create_object_data_file(part_id,0,PART_OBJECT_TYPE))
	{
		print_msg(ERROR_MSG," Error Creating part object data file\n");
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_DATA_PROTECT,SCSI_QUOTA_ERROR_ASC,SCSI_QUOTA_ERROR_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
		//FUTURE: this error type is not known
		return 1;
	}
	print_msg(dlevel,"OSD partition data file created successfully\n");
	
	//Create partition attribute pages
	if(ret=create_partition_attribute_pages_effects(part_id,NULL))
	{
		print_msg(ERROR_MSG," Error Creating partition attribute pages command \n");
		
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_DATA_PROTECT,SCSI_QUOTA_ERROR_ASC,SCSI_QUOTA_ERROR_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
		//FUTURE: this error type is not known
		if(ret==ROOT_NO_PART_OBJ_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_ROOT_QUOTAS,OSD_ATTR_NUM_ROOT_QUOTAS_PARTITION_COUNT);	
		} else if(ret==ROOT_CAPACITY_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_ROOT_INFO,OSD_ATTR_NUM_ROOT_INFO_TOTAL_CAPACITY);		
		
		}
		return 1;
	}

	print_msg(dlevel,"OSD partition attr pages created successfully\n");
	
	//Add partition to root data file 	
	add_object_to_collection_list(0,0,part_id,PART_OBJECT_TYPE);
	print_msg(dlevel,"OSD partition added to root data file\n");
	// FUTURE quota error can be here too	
	
	//Update command functions bit for command execution
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);

	print_msg(dlevel,"OSD create partition done successfully,updating command pages\n");
	//Update Current_command page
	update_current_command_page(osd_req,0);	

	//Update command functions bits related to set attr effects of command execution
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);
	
	//Build osd response as SCSI GOOD 
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,0);
	print_msg(dlevel,"OSD partition object created successfully\n");
	
	return 0;
	


}
//----------------------------------------------------------------------
int osd_create_collection(OSD_REQUEST *osd_req)
{

	OSD_CDB_CREATE_COLLECTION_COMMAND_HDR *osd_cdb_command_create_coll;
	u64 coll_object_id;int ret;
	osd_cdb_command_create_coll=(OSD_CDB_CREATE_COLLECTION_COMMAND_HDR *)&(osd_req->osd_cdb_struct);
	
	if(object_existance_test(osd_req,osd_cdb_command_create_coll->part_id,0,PART_OBJECT_TYPE))
	return 1;



	if(osd_cdb_command_create_coll->req_coll_obj_id)
	{
		//Check if collection already exists
		if(object_exists(osd_cdb_command_create_coll->part_id ,osd_cdb_command_create_coll->req_coll_obj_id,COLL_OBJECT_TYPE)|| object_exists(osd_cdb_command_create_coll->part_id ,osd_cdb_command_create_coll->req_coll_obj_id,UOBJ_OBJECT_TYPE) || osd_cdb_command_create_coll->req_coll_obj_id < 0x10000)
		{
		print_msg(ERROR_MSG,"Collection object with part_id=0x%llx coll-obj-id=0x%llx already exists or its <0x10000",osd_cdb_command_create_coll->part_id, osd_cdb_command_create_coll->req_coll_obj_id);
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
		//offset req_collection_id in CDB=24
		create_osd_key_specific_sense_data_desc(osd_req,24,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
				
		return 1;
		
		}
		coll_object_id=osd_cdb_command_create_coll->req_coll_obj_id;
	}else
	{
		//Get random Collection ID
		while(1)
		{
			coll_object_id=get_random_object_id(COLL_OBJECT_TYPE);
			if(!object_exists(osd_cdb_command_create_coll->part_id,coll_object_id,COLL_OBJECT_TYPE) && !object_exists(osd_cdb_command_create_coll->part_id,osd_cdb_command_create_coll->req_coll_obj_id,UOBJ_OBJECT_TYPE))
			break;
			
		}
	
	}
	
	osd_req->object_type=COLL_OBJECT_TYPE;
	osd_req->part_id=osd_cdb_command_create_coll->part_id;
	osd_req->object_id=coll_object_id;

	print_msg(dlevel,"OSD partition id= 0x%llx collection_object_id=0x%llx\n",osd_cdb_command_create_coll->part_id,coll_object_id);
	
	//Create collection attribute pages
	if(ret=create_uobj_coll_attribute_pages_effects(osd_cdb_command_create_coll->part_id,coll_object_id,COLL_OBJECT_TYPE))
	{
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_DATA_PROTECT,SCSI_QUOTA_ERROR_ASC,SCSI_QUOTA_ERROR_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
	
		if(ret==PART_NO_USER_OBJ_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_PART_QUOTAS,OSD_ATTR_NUM_PART_QUOTAS_OBJ_COUNT);		
		
		}
		else if(ret==ROOT_CAPACITY_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_ROOT_INFO,OSD_ATTR_NUM_ROOT_INFO_TOTAL_CAPACITY);		
		
		}
		else if(ret==PART_CAPACITY_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_PART_QUOTAS,OSD_ATTR_NUM_PART_QUOTAS_CAPACITY_QUOTA);		
		
		}
		return 1;
	}

	print_msg(dlevel,"OSD collection attr pages created successfully\n");

	//Create collection data file to store list of objects which are member of this collection
	if(create_object_data_file(osd_cdb_command_create_coll->part_id,coll_object_id,COLL_OBJECT_TYPE))
	{
	
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_DATA_PROTECT,SCSI_QUOTA_ERROR_ASC,SCSI_QUOTA_ERROR_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
		return 1;
	}
	print_msg(dlevel,"OSD collection data files created successfully\n");
	
	//Add this collection object to partition data file	
	add_object_to_collection_list(osd_cdb_command_create_coll->part_id,0,coll_object_id,COLL_OBJECT_TYPE);
	print_msg(dlevel,"OSD collection object is added to partition data file\n");

	//Update command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	print_msg(dlevel,"OSD collection object create done,updaint command page\n");
	
	//Update Current_command page
	update_current_command_page(osd_req,0);	
	
	//Update command function bits for set attr effects of command 
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);
	
	//Build SCSI response as SCSI GOOD 
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,0);
	print_msg(dlevel,"OSD collection object created successfully\n");
	
	return 0;
	


}
//----------------------------------------------------------------------
int osd_create(OSD_REQUEST *osd_req)
{
	OSD_CDB_CREATE_COMMAND_HDR *osd_cdb_command_create;
	u64 object_id;int ret,i,j;
	int obj_count;
	osd_cdb_command_create=(OSD_CDB_CREATE_COMMAND_HDR *)&(osd_req->osd_cdb_struct);
	obj_count=osd_cdb_command_create->no_of_objects;
	if(obj_count==0)	obj_count=1;
	
	print_msg(dlevel," Req Obj Id=0x%llx  Obj Count=0x%x",osd_cdb_command_create->req_obj_id,obj_count);

	//Check if part id exists 	
	if(object_existance_test(osd_req,osd_cdb_command_create->part_id,0,PART_OBJECT_TYPE))
	return 1;

	if(osd_cdb_command_create->req_obj_id)
	{
		u64 obj_id = osd_cdb_command_create->req_obj_id;
		
		for(i=0;i<obj_count;i++)
		{
			if(object_exists(osd_cdb_command_create->part_id ,obj_id+i,UOBJ_OBJECT_TYPE) ||object_exists(osd_cdb_command_create->part_id ,obj_id+i,COLL_OBJECT_TYPE) || obj_id<0x10000)
			{
				print_msg(ERROR_MSG,"Uobject object with part_id=0x%llx uobj-id=0x%llx already exists or its <0x10000",osd_cdb_command_create->part_id, obj_id);
		
				osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
				create_osd_error_id_sense_data_desc(osd_req);
				//offset req_object_id in CDB=24
				create_osd_key_specific_sense_data_desc(osd_req,24,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
				return 1;
		
			}
		
		}
		object_id=osd_cdb_command_create->req_obj_id;
	}else
	{	//get random obj-id
		while(1)
		{
			object_id=get_random_object_id(UOBJ_OBJECT_TYPE);
			for(i=0;i<obj_count;i++)
			{
				if(object_exists(osd_cdb_command_create->part_id,object_id+i,UOBJ_OBJECT_TYPE)|| object_exists(osd_cdb_command_create->part_id,object_id+i,COLL_OBJECT_TYPE))
				break;
			}
			if(i==obj_count) break;
		}
	
	}
	
	print_msg(dlevel,"OSD partition id= 0x%llx object_id=0x%llx\n",osd_cdb_command_create->part_id,object_id);
	
	
	osd_req->object_type=UOBJ_OBJECT_TYPE;
	osd_req->part_id=osd_cdb_command_create->part_id;
	osd_req->object_id=object_id;

	for(i=0;i<obj_count;i++)
	{
	
		//Create UObj attribute pages
		if(ret=create_uobj_coll_attribute_pages_effects(osd_cdb_command_create->part_id,object_id+i,UOBJ_OBJECT_TYPE))
		{
		
			print_msg(ERROR_MSG,"Error for create_uobj_coll_attribute_pages_effects for object_id=0x%llx ",object_id+i);
			osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_DATA_PROTECT,SCSI_QUOTA_ERROR_ASC,SCSI_QUOTA_ERROR_ASCQ,0);
			create_osd_error_id_sense_data_desc(osd_req);
			if(ret==PART_NO_USER_OBJ_QUOTA_EXCEED)
			{
				create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_PART_QUOTAS,OSD_ATTR_NUM_PART_QUOTAS_OBJ_COUNT);		
			
			}
			else if(ret==ROOT_CAPACITY_QUOTA_EXCEED)
			{
				create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_ROOT_INFO,OSD_ATTR_NUM_ROOT_INFO_TOTAL_CAPACITY);		
			
			}
			else if(ret==PART_CAPACITY_QUOTA_EXCEED)
			{
				create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_PART_QUOTAS,OSD_ATTR_NUM_PART_QUOTAS_CAPACITY_QUOTA);		
			
			}
			//Code to clean previously created objects
			print_msg(ERROR_MSG,"Removing already created objects from object_id=0x%llx+%d to %d   ",object_id,0,i-1);
			
			for(j=i-1;j>=0;j--)
			{
				remove_object_data_file(osd_cdb_command_create->part_id,object_id+j,UOBJ_OBJECT_TYPE);
				
				remove_object_attr_file(osd_cdb_command_create->part_id,object_id+j,UOBJ_OBJECT_TYPE);
				//FUTURE INTERNAL ERROR
			}
			return 1;
			
		
		
		}
	
		print_msg(dlevel,"OSD uobjects attr pages created successfully for obj-id=0x%llx\n",object_id+i);
		//Create object data file
		if(create_object_data_file(osd_cdb_command_create->part_id,object_id+i,UOBJ_OBJECT_TYPE))
		{
		
			print_msg(ERROR_MSG,"Error for create_object_data_file for object_id=0x%llx ",object_id+i);	
		
			osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_DATA_PROTECT,SCSI_QUOTA_ERROR_ASC,SCSI_QUOTA_ERROR_ASCQ,0);
			create_osd_error_id_sense_data_desc(osd_req);
			
			print_msg(ERROR_MSG,"Removing already created objects from object_id=0x%llx+%d to %d   ",object_id,0,i);
			
			for(j=i;j>=0;j--)
			{
				remove_object_data_file(osd_cdb_command_create->part_id,object_id+j,UOBJ_OBJECT_TYPE);
				
				remove_object_attr_file(osd_cdb_command_create->part_id,object_id+j,UOBJ_OBJECT_TYPE);
				//FUTURE INTERNAL ERROR
			}
					
			return 1;
			
		
		}
		print_msg(dlevel,"OSD uobject data file created successfully object_id=0x%llx\n",object_id+i);
		
	}
	
	for(i=0;i<obj_count;i++)
	{	//Add each object to partition data file
		add_object_to_collection_list(osd_cdb_command_create->part_id,0,object_id+i,UOBJ_OBJECT_TYPE);
		print_msg(dlevel,"OSD user object object_id=0x%llx is added to partition root data file\n",object_id+i);
	}

	//Set command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	print_msg(dlevel,"OSD user objects create done,updating command page\n");
	
	//Update Current_command page
	update_current_command_page(osd_req,0);	
	
	//Set command function bits for set attr effects of command
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);
	
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,0);
	print_msg(dlevel,"OSD uobject created successfully\n");
	
	
	return 0;
	
}
//----------------------------------------------------------------------
int osd_remove(OSD_REQUEST *osd_req)
{
	//OSD_RESPONSE_HEADER osd_response;
	OSD_CDB_REMOVE_COMMAND_HDR *osd_cdb_command_remove;
	osd_cdb_command_remove=(OSD_CDB_REMOVE_COMMAND_HDR *)&(osd_req->osd_cdb_struct);
	
	//Check object exists
	if(object_existance_test(osd_req,osd_cdb_command_remove->part_id,osd_cdb_command_remove->rem_obj_id,UOBJ_OBJECT_TYPE))
	return 1;

	osd_req->object_type=UOBJ_OBJECT_TYPE;
	osd_req->part_id=osd_cdb_command_remove->part_id;
	osd_req->object_id=osd_cdb_command_remove->rem_obj_id;

	remove_object_data_file(osd_cdb_command_remove->part_id,osd_cdb_command_remove->rem_obj_id,UOBJ_OBJECT_TYPE);
	//FUTURE: can be INTERNAL ERROR here
	remove_object_attr_file(osd_cdb_command_remove->part_id,osd_cdb_command_remove->rem_obj_id,UOBJ_OBJECT_TYPE);
	//FUTURE: Can be INTERNAL ERROR here

	remove_object_from_collection_list(osd_cdb_command_remove->part_id,0,osd_cdb_command_remove->rem_obj_id,UOBJ_OBJECT_TYPE);
	//FUTURE: Can be INTERNAL ERROR here
	
	remove_uobj_coll_attribute_pages_effects(osd_cdb_command_remove->part_id,osd_cdb_command_remove->rem_obj_id,UOBJ_OBJECT_TYPE);
	//FUTURE: Can be INTERNAL ERROR here
	
        //Set command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	print_msg(dlevel,"OSD remove user object  done,updating command page\n");
	
	//Update Current_command page
	update_current_command_page(osd_req,0);	
	
        //Set command function bits for set attr effects of command
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);
	
	//Build	OSD response as SCSI GOOD
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,0);
	print_msg(dlevel,"OSD object removed successfully\n");
	
	return 0;	

}
//---------------------------------------------------------------------

int osd_remove_partition(OSD_REQUEST *osd_req)
{
	//OSD_RESPONSE_HEADER osd_response;
	OSD_CDB_REMOVE_PARTITION_COMMAND_HDR *osd_cdb_command_remove_partition;

	osd_cdb_command_remove_partition=(OSD_CDB_REMOVE_PARTITION_COMMAND_HDR *)&(osd_req->osd_cdb_struct);
	
	if(object_existance_test(osd_req,osd_cdb_command_remove_partition->rem_part_id,0,PART_OBJECT_TYPE))
	return 1;

	osd_req->object_type=PART_OBJECT_TYPE;
	osd_req->part_id=osd_cdb_command_remove_partition->rem_part_id;
	osd_req->object_id=0;

	{
	//For checking partition empty 
	//	u8 data_buf[FILE_RW_BUFF_SIZE];
		OSD_DATA_IOV data_buf;
		osd_data_iov_alloc(&data_buf,FILE_RW_BUFF_SIZE,osd_req->data_iov_size);
		OSD_LIST_COMMAND_REPLY command_reply;
		command_reply.coll_part_id=osd_cdb_command_remove_partition->rem_part_id;
		command_reply.coll_id=0;
		command_reply.object_type=PART_OBJECT_TYPE;
		command_reply.continuation_id=0;
		command_reply.initial_data_offset=0;
		command_reply.data_buf=&data_buf;
		command_reply.data_buf_size=FILE_RW_BUFF_SIZE;
		command_reply.object_count=0;
	
		if(list_objects_from_collection_list(&command_reply))
		{
			print_msg(ERROR_MSG,"Error list_objects_from_collection_list\n");
			PRINT_OSD_LIST_COMMAND_REPLY_PTR((&command_reply));
		//	return 1;
		}
		//print_objects_from_collection_list(&command_reply);
	
		if(command_reply.object_count)
		{
			print_msg(dlevel,"Partition object with part_id=0x%llx is not empty",osd_cdb_command_remove_partition->rem_part_id); 
			osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_PARTITION_NOT_EMPTY_ASC,SCSI_PARTITION_NOT_EMPTY_ASCQ,0);
			create_osd_error_id_sense_data_desc(osd_req);
			osd_data_iov_free(&data_buf);
			return 1; 
		}
		osd_data_iov_free(&data_buf);
	
	}

	
	remove_object_attr_file(osd_cdb_command_remove_partition->rem_part_id,0,PART_OBJECT_TYPE);
	//FUTURE:Can be INTERNAL ERROR here
	remove_object_from_collection_list(0,0,osd_cdb_command_remove_partition->rem_part_id,PART_OBJECT_TYPE);
	//FUTURE:Can be INTERNAL ERROR here
	remove_partition_attribute_pages_effects(osd_cdb_command_remove_partition->rem_part_id);
	//FUTURE:Can be INTERNAL ERROR here
	
	{	//For testing purpose 
		//	u8 data_buf[FILE_RW_BUFF_SIZE];
		OSD_DATA_IOV data_buf;
		osd_data_iov_alloc(&data_buf,FILE_RW_BUFF_SIZE,osd_req->data_iov_size);
	
		OSD_LIST_COMMAND_REPLY command_reply;
		command_reply.coll_part_id=0;
		command_reply.coll_id=0;
		command_reply.object_type=ROOT_OBJECT_TYPE;
		command_reply.continuation_id=0;
		command_reply.initial_data_offset=0;
		command_reply.data_buf=&data_buf;
		command_reply.data_buf_size=FILE_RW_BUFF_SIZE;
		command_reply.object_count=0;
		
			if(list_objects_from_collection_list(&command_reply))
			{
				print_msg(ERROR_MSG,"Error list_objects_from_collection_list\n");
				PRINT_OSD_LIST_COMMAND_REPLY_PTR((&command_reply));
			//	return 1;
			}
		print_objects_from_collection_list(&command_reply);
	
		osd_data_iov_free(&data_buf);	
	}

        //Set command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	print_msg(dlevel,"OSD remover partition object done,updating command page\n");
	
	//Update Current_command page
	update_current_command_page(osd_req,0);	
	
        //Set command function bits for set attr effects of command
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);
	
        //Build OSD response as SCSI GOOD
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,0);
	print_msg(dlevel,"OSD partition object removed successfully\n");
	return 0;	
	
}

//---------------------------------------------------------------------
int osd_remove_collection(OSD_REQUEST *osd_req)
{
	OSD_CDB_REMOVE_COLLECTION_COMMAND_HDR *osd_cdb_command_remove_collection;
	u8 FCR_BIT;
	osd_cdb_command_remove_collection=(OSD_CDB_REMOVE_COLLECTION_COMMAND_HDR *)&(osd_req->osd_cdb_struct);

	if(object_existance_test(osd_req,osd_cdb_command_remove_collection->part_id,osd_cdb_command_remove_collection->rem_coll_obj_id,COLL_OBJECT_TYPE))
	return 1;

	
	osd_req->object_type=COLL_OBJECT_TYPE;
	osd_req->part_id=osd_cdb_command_remove_collection->part_id;
	osd_req->object_id=osd_cdb_command_remove_collection->rem_coll_obj_id;

	FCR_BIT=osd_cdb_command_remove_collection->resvd[11];
	FCR_BIT&=1;
	print_msg(dlevel,"FCR Bit= %d",FCR_BIT);
	
	
	//FUTURE HANDLE WHEN FCR_BIT=1  REFER 6.19
	if(!FCR_BIT)
	{
	//	Check collection object empty or not
	//	u8 data_buf[FILE_RW_BUFF_SIZE];
		OSD_DATA_IOV data_buf;
		osd_data_iov_alloc(&data_buf,FILE_RW_BUFF_SIZE,osd_req->data_iov_size);
		OSD_LIST_COMMAND_REPLY command_reply;
		command_reply.coll_part_id=osd_cdb_command_remove_collection->part_id;
		command_reply.coll_id=osd_cdb_command_remove_collection->rem_coll_obj_id;
		command_reply.object_type=COLL_OBJECT_TYPE;
		command_reply.continuation_id=0;
		command_reply.initial_data_offset=0;
		command_reply.data_buf=&data_buf;
		command_reply.data_buf_size=FILE_RW_BUFF_SIZE;
		command_reply.object_count=0;
		
		if(list_objects_from_collection_list(&command_reply))
		{
			print_msg(ERROR_MSG,"Error list_objects_from_collection_list\n");
			PRINT_OSD_LIST_COMMAND_REPLY_PTR((&command_reply));
			//	return 1;
		}
		//print_objects_from_collection_list(&command_reply);
		
		if(command_reply.object_count)
		{
			print_msg(dlevel,"Collection object with part_id=0x%llx Coll_obj_id=0x%llx is not empty", osd_cdb_command_remove_collection->part_id,osd_cdb_command_remove_collection->rem_coll_obj_id); 
			osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_PARTITION_NOT_EMPTY_ASC,SCSI_PARTITION_NOT_EMPTY_ASCQ,0);
			create_osd_error_id_sense_data_desc(osd_req);
			osd_data_iov_free(&data_buf);
			return 1; 
		}
		osd_data_iov_free(&data_buf);		
	
	}	

	remove_object_data_file(osd_cdb_command_remove_collection->part_id,osd_cdb_command_remove_collection->rem_coll_obj_id,COLL_OBJECT_TYPE);
	//FUTURE: Can be INTERNAL ERROR here
	remove_object_attr_file(osd_cdb_command_remove_collection->part_id,osd_cdb_command_remove_collection->rem_coll_obj_id,COLL_OBJECT_TYPE);
	//FUTURE: Can be INTERNAL ERROR here

	remove_object_from_collection_list(osd_cdb_command_remove_collection->part_id,0,osd_cdb_command_remove_collection->rem_coll_obj_id,COLL_OBJECT_TYPE);
	//FUTURE: Can be INTERNAL ERROR here
	
	remove_uobj_coll_attribute_pages_effects(osd_cdb_command_remove_collection->part_id,osd_cdb_command_remove_collection->rem_coll_obj_id,COLL_OBJECT_TYPE);

        //Set command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	print_msg(dlevel,"OSD remove collection object done,updating command page\n");
	
	//Update Current_command page
	update_current_command_page(osd_req,0);	

        //Set command function bits for set attr effects of command
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);
	
        //Build OSD response as SCSI GOOD
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,0);
	print_msg(dlevel,"OSD object removed successfully\n");
	
	return 0;
}
//----------------------------------------------------------------------
int osd_read(OSD_REQUEST *osd_req)
{

	OSD_CDB_RW_COMMAND_HDR *osd_cdb_rw_hdr;
	osd_cdb_rw_hdr=(OSD_CDB_RW_COMMAND_HDR *)&(osd_req->osd_cdb_struct);

	u64 user_obj_logical_len;
	u64 read_size,read_len;

	//Check if object exists
	if(object_existance_test(osd_req,osd_cdb_rw_hdr->part_id,osd_cdb_rw_hdr->object_id,UOBJ_OBJECT_TYPE))
	return 1;

	osd_req->object_type=UOBJ_OBJECT_TYPE;
	osd_req->part_id=osd_cdb_rw_hdr->part_id;
	osd_req->object_id=osd_cdb_rw_hdr->object_id;
	
	get_page_attr(osd_req->part_id,osd_req->object_id,OSD_PAGE_NUM_UOBJ_INFO,OSD_ATTR_NUM_UOBJ_INFO_OBJ_LENGTH,&(user_obj_logical_len),UOBJ_OBJECT_TYPE);
	
	if(user_obj_logical_len <= osd_cdb_rw_hdr->starting_byte_offset)
	{
	
		print_msg(ERROR_MSG,"Error user_obj_logical_len=0x%llx Read starting_byte_offset=0x%llx\n",user_obj_logical_len,osd_cdb_rw_hdr->starting_byte_offset);
			
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,osd_req->osd_response_header.data_len);
		create_osd_error_id_sense_data_desc(osd_req);
	
		//offset req_object_id in CDB=24
		create_osd_key_specific_sense_data_desc(osd_req,44,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
		return 1;
		//check status
	}
	
	if(osd_cdb_rw_hdr->starting_byte_offset+osd_cdb_rw_hdr->length > user_obj_logical_len)
	{
		read_size= user_obj_logical_len-osd_cdb_rw_hdr->starting_byte_offset;
	}
	else
	read_size=osd_cdb_rw_hdr->length;
	
	read_len=perform_rw_operation(osd_req->part_id,osd_req->object_id,&(osd_req->data_out),0,osd_cdb_rw_hdr->starting_byte_offset,read_size,1);
	print_msg(dlevel,"Data size read= 0x%llx",read_len);
	
	//FUTURE there can be internal error while reading here
	
	
        //Set command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	print_msg(dlevel,"OSD read object done,updating command page\n");
	
	//Update Current_command page
	update_current_command_page(osd_req,0);	

	read_attribute_pages_effects(osd_req->part_id,osd_req->object_id);

        //Set command function bits for set attr effects of command
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);

        //Build OSD response as SCSI GOOD
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,read_len);
	
	print_msg(dlevel,"OSD read object done successfully\n");
	
	return 0;	
}
//----------------------------------------------------------------------
int osd_write_append(OSD_REQUEST *osd_req)
{

	OSD_CDB_RW_COMMAND_HDR *osd_cdb_rw_hdr;
	osd_cdb_rw_hdr=(OSD_CDB_RW_COMMAND_HDR *)&(osd_req->osd_cdb_struct);
	int ret;
	u16 service_action;
	u64 written_len;
	WRITE_AFFECTED_ATTRS write_affected_attrs;

	if(object_existance_test(osd_req,osd_cdb_rw_hdr->part_id,osd_cdb_rw_hdr->object_id,UOBJ_OBJECT_TYPE))
	return 1;

	osd_req->object_type=UOBJ_OBJECT_TYPE;
	osd_req->part_id=osd_cdb_rw_hdr->part_id;
	osd_req->object_id=osd_cdb_rw_hdr->object_id;
	
	service_action=cmnd_opcode(osd_req);
	
	write_affected_attrs.object_id=osd_cdb_rw_hdr->object_id;
	write_affected_attrs.part_id=osd_cdb_rw_hdr->part_id;
	if(service_action!=OSD_CMD_APPEND) //it can be WRITE or CREATE_AND_WRITE
	write_affected_attrs.starting_byte_offset=osd_cdb_rw_hdr->starting_byte_offset;
	write_affected_attrs.length=osd_cdb_rw_hdr->length;
	
	if(ret=get_and_check_write_affected_attributes(service_action,&write_affected_attrs))
	{
		print_msg(ERROR_MSG,"Error in get_and_check_write_affected_attributes");
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_DATA_PROTECT,SCSI_QUOTA_ERROR_ASC,SCSI_QUOTA_ERROR_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
		//FUTURE: this error type is not known
		if(ret==UOBJ_LENGTH_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_UOBJ_QUOTAS, OSD_ATTR_NUM_UOBJ_QUOTAS_MAX_OBJ_LEN);
		//OSD_REQUEST *osd_req,u16 page_no,u16 attr_no);
		} else if(ret==PART_CAPACITY_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_PART_QUOTAS,OSD_ATTR_NUM_PART_QUOTAS_CAPACITY_QUOTA);
		
		} else if(ret==ROOT_CAPACITY_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_ROOT_INFO,OSD_ATTR_NUM_ROOT_INFO_TOTAL_CAPACITY);		
		}
		return 1;
	}
	print_msg(dlevel,"get_and_check_write_affected_attributes done");
	
	written_len=perform_rw_operation(write_affected_attrs.part_id,write_affected_attrs.object_id,&(osd_req->data_in),0,write_affected_attrs.starting_byte_offset,write_affected_attrs.length,0);
	print_msg(dlevel,"Data size written= 0x%llx",written_len);
	
        //Set command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
		
	set_write_affected_attributes(&write_affected_attrs);
	print_msg(dlevel,"set_write_affected_attributes done");

	print_msg(dlevel,"OSD write/append object done,updating command page\n");
	
	//Update Current_command page
	update_current_command_page(osd_req,0);	

	 //Set command function bits for set attr effects of command
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);
	
        //Build OSD response as SCSI GOOD
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,0);
	
	print_msg(dlevel,"OSD write/append object done successfully\n");
	
	return 0;	
}
//---------------------------------------------------------------------------
int osd_list(OSD_REQUEST *osd_req)
{	

	OSD_CDB_LIST_COMMAND_HDR *osd_cdb_list_hdr;
	OSD_LIST_COMMAND_REPLY command_reply;
	osd_cdb_list_hdr=(OSD_CDB_LIST_COMMAND_HDR *)&(osd_req->osd_cdb_struct);

  	print_msg(dlevel,"part_id=0x%llx ",osd_cdb_list_hdr->part_id);
    	//We dont consider SORT Bit for list reply
	if(osd_cdb_list_hdr->part_id)
	{
		if(object_existance_test(osd_req,osd_cdb_list_hdr->part_id,0,PART_OBJECT_TYPE))
		return 1;
	}
	osd_req->part_id=osd_cdb_list_hdr->part_id;
	osd_req->object_id=0;
	
	if(osd_cdb_list_hdr->part_id ==0 )
		osd_req->object_type=ROOT_OBJECT_TYPE;
	else
		osd_req->object_type=PART_OBJECT_TYPE;
	
	u32 read_len=0;
	u64 continuation_id=0;
	
	/**Use internal structure to get list of object */
	command_reply.final=0;
	command_reply.coll_part_id=osd_req->part_id;
	command_reply.coll_id=osd_req->object_id;
	command_reply.object_type=osd_req->object_type;
	command_reply.continuation_id=osd_cdb_list_hdr->initial_obj_id;
	command_reply.initial_data_offset=24;
	command_reply.data_buf=&(osd_req->data_out);
	command_reply.data_buf_size=osd_cdb_list_hdr->allocation_length-24;//get_alloction len
	command_reply.object_count=0;
		
	if(list_objects_from_collection_list(&command_reply))
	{	//This is internal error
		print_msg(ERROR_MSG,"Error list_objects_from_collection_list\n");
		PRINT_OSD_LIST_COMMAND_REPLY_PTR((&command_reply));
                osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
                create_osd_error_id_sense_data_desc(osd_req);
                //offset part_id in CDB=16
                create_osd_key_specific_sense_data_desc(osd_req,16,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
                return 1;
	//	return 1;
	}
	print_objects_from_collection_list(&command_reply);
	read_len =+(command_reply.object_count*8);
	
	//If all object list is not done, set continuation ID
	if(! command_reply.final)
	{
		if(command_reply.object_count)
		{
			if(iov_bufcpy(command_reply.data_buf,(u8 *) &continuation_id,(command_reply.object_count-1)*8+24,8,FROM_IOV_BUF)!=8)
			{
				print_msg(ERROR_MSG,"Internal Error unable to get continuation id");
				continuation_id=0;
			
				PRINT_OSD_LIST_COMMAND_REPLY_PTR((&command_reply));
                		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
                		create_osd_error_id_sense_data_desc(osd_req);
                		//offset part_id in CDB=16
                		create_osd_key_specific_sense_data_desc(osd_req,16,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
                		return 1;
			}
		}
	}

	//Set various fields as LIST reply according to T10 OSD-2 specification 6.13 
	u64 add_len=read_len+16; // 16 for bytes 8-23 of table 66 T10 OSD-2 specification 6.13 

	u32 list_id=0;//we dont care this value
	//LSTCHG bit 0 
	u8 root_bit=0;
	if(osd_req->object_type==ROOT_OBJECT_TYPE)
	root_bit=1;

	add_len=cpu_to_be64(add_len);

	if(iov_bufcpy(command_reply.data_buf,(u8 *) &add_len,0,8,TO_IOV_BUF)!=8)
	{
		print_msg(ERROR_MSG,"Internal Error unable to copy add_len");
		return 1; 
	}

	
	print_msg(dlevel,"Call to iov_bufcpy to set continuation_id=0x%llx",be64_to_cpu(continuation_id));
	//continuation_id=cpu_to_be64(continuation_id);
	if(iov_bufcpy(command_reply.data_buf,(u8 *) &continuation_id,8,8,TO_IOV_BUF)!=8)
	{
		print_msg(ERROR_MSG,"Internal Error unable to copy continuation_id");
		return 1; 
	}

	//memcpy(ptr+8,&continuation_id,8);
	list_id=cpu_to_be32(list_id);
	print_msg(dlevel,"Call to iov_bufcpy to set list_id");
	if(iov_bufcpy(command_reply.data_buf,(u8 *) &list_id,16,4,TO_IOV_BUF)!=4)
	{
		print_msg(ERROR_MSG,"Internal Error unable to copy list_id");
		return 1; 
	}
		
	//memcpy(ptr+16,&add_len,8);
	//memcpy(ptr+23,&root_bit,1);
	print_msg(dlevel,"Call to iov_bufcpy to set list_id");
	if(iov_bufcpy(command_reply.data_buf,(u8 *)&root_bit,23,1,TO_IOV_BUF)!=1)
	{
		print_msg(ERROR_MSG,"Internal Error unable to copy root_bit");
		return 1; 
	}
	
        //Set command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	print_msg(dlevel,"OSD list operation done,updating related attribute pages and  command page\n");
	
	list_attribute_pages_effects(osd_req->part_id,osd_req->object_id);
	//Update Current_command page
	update_current_command_page(osd_req,0);	

         //Set command function bits for set attr effects of command
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);

       //Build OSD response as SCSI GOOD
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,read_len+24);
	
	print_msg(dlevel,"OSD list object done successfully\n");
	
	return 0;	
}
//---------------------------------------------------------------------------
int osd_list_collection(OSD_REQUEST *osd_req)
{

	OSD_CDB_LIST_COLLECTION_COMMAND_HDR *osd_cdb_list_coll_hdr;
	OSD_LIST_COMMAND_REPLY command_reply;
	osd_cdb_list_coll_hdr=(OSD_CDB_LIST_COLLECTION_COMMAND_HDR *)&(osd_req->osd_cdb_struct);

  	print_msg(dlevel,"Enter");
  	print_msg(dlevel,"part_id=0x%llx Collection ID=0x%llx",osd_cdb_list_coll_hdr->part_id,osd_cdb_list_coll_hdr->coll_id);

	if(object_existance_test(osd_req,osd_cdb_list_coll_hdr->part_id,osd_cdb_list_coll_hdr->coll_id,COLL_OBJECT_TYPE))
	return 1;
	
	osd_req->part_id=osd_cdb_list_coll_hdr->part_id;
	osd_req->object_id=osd_cdb_list_coll_hdr->coll_id;;
	
	osd_req->object_type=COLL_OBJECT_TYPE;
	
	u32 read_len=0;
	u64 continuation_id=0;
	
	command_reply.final=0;
	command_reply.coll_part_id=osd_req->part_id;
	command_reply.coll_id=osd_req->object_id;
	command_reply.object_type=osd_req->object_type;
	command_reply.continuation_id=osd_cdb_list_coll_hdr->initial_obj_id;
	command_reply.initial_data_offset=24;
	command_reply.data_buf=&(osd_req->data_out);
	command_reply.data_buf_size=osd_cdb_list_coll_hdr->allocation_length-24;//get_allocation len
	command_reply.object_count=0;
		
	if(list_objects_from_collection_list(&command_reply))
	{
		print_msg(ERROR_MSG,"Error list_objects_from_collection_list\n");
		PRINT_OSD_LIST_COMMAND_REPLY_PTR((&command_reply));
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
                create_osd_error_id_sense_data_desc(osd_req);
                //offset req_object_id in CDB=24
                create_osd_key_specific_sense_data_desc(osd_req,24,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
		return 1;
	}
	print_objects_from_collection_list(&command_reply);
			
	read_len =+(command_reply.object_count*8);
		
	if(! command_reply.final)
	{
		if(command_reply.object_count)
		{
			
			if(iov_bufcpy(command_reply.data_buf,(u8 *) &continuation_id,(command_reply.object_count-1)*8+24,8,FROM_IOV_BUF)!=8)
			{
				print_msg(ERROR_MSG,"Internal Error unable to get continuation id");
				continuation_id=0;
				osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
                		create_osd_error_id_sense_data_desc(osd_req);
                		//offset req_object_id in CDB=24
                		create_osd_key_specific_sense_data_desc(osd_req,24,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
				return 1; 
			}
	
		}
	}

	//Set various fields as LIST reply according to T10 OSD-2 specification 6.14 
	
	u64 add_len=read_len+16; // 16 for bytes 8-23 of table 66 T10 OSD-2 specification 6.14 
	
	u32 list_id=0;//we dont care this value
	//LSTCHG bit 0 
	u8 root_bit=0;
	if(osd_req->object_type==ROOT_OBJECT_TYPE) //For Historical purpose :)
		root_bit=1;
	add_len=cpu_to_be64(add_len);
	print_msg(dlevel,"Call to iov_bufcpy to set add_len");
	if(iov_bufcpy(command_reply.data_buf,(u8 *) &add_len,0,8,TO_IOV_BUF)!=8)
	{
		print_msg(ERROR_MSG,"Internal Error unable to copy add_len");
		return 1; 
	}
	
	print_msg(dlevel,"Call to iov_bufcpy to set continuation_id=0x%llx",be64_to_cpu(continuation_id));
	//continuation_id=cpu_to_be64(continuation_id);
	if(iov_bufcpy(command_reply.data_buf,(u8 *) &continuation_id,8,8,TO_IOV_BUF)!=8)
	{
		print_msg(ERROR_MSG,"Internal Error unable to copy continuation_id");
		return 1; 
	}

	//memcpy(ptr+8,&continuation_id,8);

	list_id=cpu_to_be32(list_id);
	print_msg(dlevel,"Call to iov_bufcpy to set list_id");
	if(iov_bufcpy(command_reply.data_buf,(u8 *) &list_id,16,4,TO_IOV_BUF)!=4)
	{
		print_msg(ERROR_MSG,"Internal Error unable to copy list_id");
		return 1; 
	}


	//memcpy(ptr+16,&add_len,8);
	//memcpy(ptr+23,&root_bit,1);
	print_msg(dlevel,"Call to iov_bufcpy to set list_id");
	if(iov_bufcpy(command_reply.data_buf,(u8 *)&root_bit,23,1,TO_IOV_BUF)!=1)
	{
		print_msg(ERROR_MSG,"Internal Error unable to copy root_bit");
		return 1; 
	}

        //Set command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	print_msg(dlevel,"OSD list operation done,updating related attribute pages and  command page\n");
	
	list_attribute_pages_effects(osd_req->part_id,osd_req->object_id);
	//Update Current_command page
	update_current_command_page(osd_req,0);	
	
        //Set command function bits for set attr effects of command
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);
	
	//Build OSD response as SCSI GOOD
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,read_len+24);
	
	print_msg(dlevel,"OSD list object done successfully\n");
	
	return 0;	

}
//---------------------------------------------------------------------------

int osd_get_set_attr(OSD_REQUEST *osd_req)
{
	OSD_CDB_ANY_COMMAND_HDR *osd_any_command;
	osd_any_command=(OSD_CDB_ANY_COMMAND_HDR *)&(osd_req->osd_cdb_struct);

	osd_req->part_id=osd_any_command->part_id;
	osd_req->object_id=osd_any_command->object_id;
	
	if(osd_req->object_id !=0)
		osd_req->object_type=UOBJ_OBJECT_TYPE;
	else if (osd_req->part_id !=0)	
		osd_req->object_type=PART_OBJECT_TYPE;
	else
		osd_req->object_type=ROOT_OBJECT_TYPE;	
	
	if(object_existance_test(osd_req,osd_req->part_id,osd_req->object_id,UOBJ_OBJECT_TYPE))
	return 1;

	//As command processing of get/set attribute will be done later as part of general get/set attrbiute processing
	//Set command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	
	update_current_command_page(osd_req,0);	

	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);

	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,0);
	print_msg(dlevel,"OSD get/set attr ");
	
	return 0;

}
//---------------------------------------------------------------------------

int osd_create_write(OSD_REQUEST *osd_req)
{

	OSD_CDB_CREATE_WRITE_COMMAND_HDR *osd_create_write_command;
	u64 object_id; int ret;
	osd_create_write_command=(OSD_CDB_CREATE_WRITE_COMMAND_HDR *)&(osd_req->osd_cdb_struct);
	
	//Code next onwards is copy of code from Create command and Write command
	
	if(object_existance_test(osd_req,osd_create_write_command->part_id,0,PART_OBJECT_TYPE))
	return 1;

	//Create Single object
       if(osd_create_write_command->req_obj_id)
	{
		u64 obj_id = osd_create_write_command->req_obj_id;
		
		if(object_exists(osd_create_write_command->part_id ,obj_id,UOBJ_OBJECT_TYPE) || obj_id < 0x10000)
		{
		print_msg(ERROR_MSG,"User object with part_id=0x%llx obj-id=0x%llx does  exists or its <0x10000",osd_create_write_command->part_id,osd_create_write_command->req_obj_id);
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
	
		//offset req_object_id in CDB=24
		create_osd_key_specific_sense_data_desc(osd_req,24,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
		
		return 1;
		
		}
		object_id=osd_create_write_command->req_obj_id;
	}
	else
	{	//Get random object-id 
		while(1)
		{
			object_id=get_random_object_id(UOBJ_OBJECT_TYPE);
			if(!object_exists(osd_create_write_command->part_id,object_id,UOBJ_OBJECT_TYPE))
			break;
		}
	
	}
	
	print_msg(dlevel,"OSD partition id= 0x%llx object_id=0x%llx\n",osd_create_write_command->part_id,object_id);
	
	osd_req->object_type=UOBJ_OBJECT_TYPE;
	osd_req->part_id=osd_create_write_command->part_id;
	osd_req->object_id=object_id;

	//Create object attribute pages
	if(ret=create_uobj_coll_attribute_pages_effects(osd_create_write_command->part_id,object_id,UOBJ_OBJECT_TYPE))
	{
		
		print_msg(ERROR_MSG,"Error for create_uobj_coll_attribute_pages_effects for object_id=0x%llx ",object_id);
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_DATA_PROTECT,SCSI_QUOTA_ERROR_ASC,SCSI_QUOTA_ERROR_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
		if(ret==PART_NO_USER_OBJ_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_PART_QUOTAS,OSD_ATTR_NUM_PART_QUOTAS_OBJ_COUNT);		
			
		}
		else if(ret==ROOT_CAPACITY_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_ROOT_INFO,OSD_ATTR_NUM_ROOT_INFO_TOTAL_CAPACITY);		
		}
		else if(ret==PART_CAPACITY_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_PART_QUOTAS,OSD_ATTR_NUM_PART_QUOTAS_CAPACITY_QUOTA);		
			
		}

		
		return 1;
			
		
		
	}
	
	print_msg(dlevel,"OSD uobject attr pages created successfully for obj-id=0x%llx\n",object_id);

	//Create Object data file	
	if(create_object_data_file(osd_create_write_command->part_id,object_id,UOBJ_OBJECT_TYPE))
	{
		
		print_msg(ERROR_MSG,"Error for create_object_data_file for object_id=0x%llx ",object_id);	
		
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_DATA_PROTECT,SCSI_QUOTA_ERROR_ASC,SCSI_QUOTA_ERROR_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
		return 1;
	}
	print_msg(dlevel,"OSD uobject data file created successfully object_id=0x%llx\n",object_id);
	

	//Add object to partition data file
	add_object_to_collection_list(osd_create_write_command->part_id,0,object_id,UOBJ_OBJECT_TYPE);
	print_msg(dlevel,"OSD user object object_id=0x%llx is added to partition root data file\n",object_id);
	
	
	print_msg(dlevel,"For OSD CREATE_AND_WRITE command user object create done for object_id=0x%llx now going to call osd_write to WRITE data\n",object_id);
	
	//Write this object code
	//Need to send just object_id field
	osd_create_write_command->req_obj_id=object_id;
	
	
	return osd_write_append(osd_req);
	
}
//---------------------------------------------------------------------------
int osd_add_to_collection(OSD_REQUEST *osd_req)
{

	OSD_CDB_ADD_TO_COLLECTION_COMMAND_HDR *osd_cdb_command_add_to_coll;
	int ret;
	
	osd_cdb_command_add_to_coll=(OSD_CDB_ADD_TO_COLLECTION_COMMAND_HDR *)&(osd_req->osd_cdb_struct);
	
	//Check if collection exists
	if(object_existance_test(osd_req,osd_cdb_command_add_to_coll->part_id,osd_cdb_command_add_to_coll->coll_obj_id,COLL_OBJECT_TYPE))
	return 1;

	//Check if user object exists
	if((!object_exists(osd_cdb_command_add_to_coll->part_id ,osd_cdb_command_add_to_coll->obj_id,UOBJ_OBJECT_TYPE)) ||    is_uobject_member_of_collection(osd_cdb_command_add_to_coll->part_id,osd_cdb_command_add_to_coll->obj_id,osd_cdb_command_add_to_coll->coll_obj_id))
	{
		print_msg(ERROR_MSG,"User object with part_id=0x%llx obj-id=0x%llx does not exists or is already member of collection 0x%llx",osd_cdb_command_add_to_coll->part_id,osd_cdb_command_add_to_coll->obj_id,osd_cdb_command_add_to_coll->coll_obj_id);


		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
		//offset req_collection_id in CDB=24
		create_osd_key_specific_sense_data_desc(osd_req,36,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
				
		return 1;
		
	}
		
	osd_req->object_type=COLL_OBJECT_TYPE;
	osd_req->part_id=osd_cdb_command_add_to_coll->part_id;
	osd_req->object_id=osd_cdb_command_add_to_coll->coll_obj_id;

	print_msg(dlevel,"OSD partition id= 0x%llx collection_object_id=0x%llx Object to add=0x%llx \n",osd_cdb_command_add_to_coll->part_id,osd_cdb_command_add_to_coll->coll_obj_id,osd_cdb_command_add_to_coll->obj_id);
	
	//Add object to collection and update related attribute pages
	if(ret=add_to_coll_attribute_pages_effects(osd_cdb_command_add_to_coll->part_id,osd_cdb_command_add_to_coll->coll_obj_id,osd_cdb_command_add_to_coll->obj_id))
	{

		print_msg(ERROR_MSG,"Error in add_to_coll_attribute_pages_effects");
		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_DATA_PROTECT,SCSI_QUOTA_ERROR_ASC,SCSI_QUOTA_ERROR_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
		//FUTURE: this error type is not known
		
		if(ret==PART_CAPACITY_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_PART_QUOTAS,OSD_ATTR_NUM_PART_QUOTAS_CAPACITY_QUOTA);
		
		} else if(ret==ROOT_CAPACITY_QUOTA_EXCEED)
		{
		create_osd_attr_id_sense_data_desc(osd_req,OSD_PAGE_NUM_ROOT_INFO,OSD_ATTR_NUM_ROOT_INFO_TOTAL_CAPACITY);		
		}
		return 1;
	}

	print_msg(dlevel,"OSD add object to collection done,updating command page\n");
	//Set command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	
	//Update Current_command page
	update_current_command_page(osd_req,0);	
	//update write side effects here
	//read_attribute_pages_effects(osd_req->part_id,osd_req->object_id);

         //Set command function bits for set attr effects of command
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);
	
        //Build OSD response as SCSI GOOD
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,0);
	
	print_msg(dlevel,"OSD add object to collection done successfully\n");
	return 0;	
}

//---------------------------------------------------------------------------
int osd_remove_from_collection(OSD_REQUEST *osd_req)
{

	OSD_CDB_REMOVE_FROM_COLLECTION_COMMAND_HDR *osd_cdb_command_remove_from_coll;
	int ret;
	
	osd_cdb_command_remove_from_coll=(OSD_CDB_REMOVE_FROM_COLLECTION_COMMAND_HDR *)&(osd_req->osd_cdb_struct);
	

	if(object_existance_test(osd_req,osd_cdb_command_remove_from_coll->part_id,osd_cdb_command_remove_from_coll->coll_obj_id,COLL_OBJECT_TYPE))
	return 1;

	
	if((!object_exists(osd_cdb_command_remove_from_coll->part_id ,osd_cdb_command_remove_from_coll->obj_id,UOBJ_OBJECT_TYPE)) )
	{
		print_msg(ERROR_MSG,"User object with part_id=0x%llx obj-id=0x%llx does not exists for collection 0x%llx",osd_cdb_command_remove_from_coll->part_id,osd_cdb_command_remove_from_coll->obj_id,osd_cdb_command_remove_from_coll->coll_obj_id);

		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
		//offset obj_id in CDB=36
		create_osd_key_specific_sense_data_desc(osd_req,36,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
				
		return 1;
	
	}
		

	//Condition of is_uobject_member_of_collection is not checked here since that can create overhead as in next step while removing we are searching object and removing

	osd_req->object_type=COLL_OBJECT_TYPE;
	osd_req->part_id=osd_cdb_command_remove_from_coll->part_id;
	osd_req->object_id=osd_cdb_command_remove_from_coll->coll_obj_id;

	print_msg(dlevel,"OSD partition id= 0x%llx collection_object_id=0x%llx Object to remove=0x%llx \n",osd_cdb_command_remove_from_coll->part_id,osd_cdb_command_remove_from_coll->coll_obj_id,osd_cdb_command_remove_from_coll->obj_id);
	
	//Remove object from collection and update related pages
	if(ret=remove_from_coll_attribute_pages_effects(osd_cdb_command_remove_from_coll->part_id,osd_cdb_command_remove_from_coll->coll_obj_id,osd_cdb_command_remove_from_coll->obj_id))
	{

		print_msg(ERROR_MSG,"Error in remove_from_coll_attribute_pages_effects");
	
		print_msg(ERROR_MSG,"User object with part_id=0x%llx obj-id=0x%llx does not exists for collection 0x%llx",osd_cdb_command_remove_from_coll->part_id,osd_cdb_command_remove_from_coll->obj_id,osd_cdb_command_remove_from_coll->coll_obj_id);

		osd_req->osd_response_header=build_osd_response_header(SCSI_CHECK_CONDITION,SCSI_SENSE_KEY_ILLEGAL_REQUEST,SCSI_INVALID_FIELD_IN_CDB_ASC,SCSI_INVALID_FIELD_IN_CDB_ASCQ,0);
		create_osd_error_id_sense_data_desc(osd_req);
		//offset obj_id in CDB=36
		create_osd_key_specific_sense_data_desc(osd_req,36,OSD_SENSE_KEY_SPF_CDB_FIELD_FLAG);
				
		return 1;
			
	}

	print_msg(dlevel,"OSD remove object from  collection done,updating command page\n");
	
	//Set command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
		
	//Update Current_command page
	update_current_command_page(osd_req,0);	
	//update write side effects here
	//read_attribute_pages_effects(osd_req->part_id,osd_req->object_id);

         //Set command function bits for set attr effects of command
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);
	
        //Build OSD response as SCSI GOOD
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,0);
	
	print_msg(dlevel,"OSD add object to collection done successfully\n");
	
	return 0;	
}
//---------------------------------------------------------------------------
int osd_other_command(OSD_REQUEST *osd_req)
{
	OSD_CDB_ANY_COMMAND_HDR *osd_any_command;

	osd_any_command=(OSD_CDB_ANY_COMMAND_HDR *)&(osd_req->osd_cdb_struct);
	
	osd_req->part_id=osd_any_command->part_id;
	osd_req->object_id=osd_any_command->object_id;

	if(object_existance_test(osd_req,osd_req->part_id,osd_req->object_id,UOBJ_OBJECT_TYPE))
	return 1;
	
	//FUTURE : no error handling as of now....in future may be
	switch(cmnd_opcode(osd_req))
	{
		case OSD_CMD_FLUSH:
				osd_req->object_type=UOBJ_OBJECT_TYPE;
				break;
		case OSD_CMD_FLUSH_PARTITION:	
				osd_req->object_type=PART_OBJECT_TYPE;
				break;
		case OSD_CMD_FLUSH_OSD:	
				osd_req->object_type=ROOT_OBJECT_TYPE;
				break;
	}

	//Set command function bits for command completion
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_BYTE,COMMAND_FUNCTIONS_COMMAND_BIT);
	update_current_command_page(osd_req,0);	
	set_command_functions_bit(osd_req,COMMAND_FUNCTIONS_COMMAND_RESULT_SET_ATTR,COMMAND_FUNCTIONS_IMP_ST_ATT_BIT);
	osd_req->osd_response_header=build_osd_response_header(SCSI_GOOD,0,0,0,0);
	print_msg(dlevel,"OSD any command done");
	
	return 0;
}
//---------------------------------------------------------------------------
