/**
 *                      ==================
 *                      IITK OSD Simulator
 *                      ==================
 *   An Implementation of Simulator for Object based storage device(OSD)
 *
 *   Copyright (C) 2007-2008 CSE IIT KANPUR
 *
 *   Contact: Department of Computer Science and Engineering
 *            Indian Institute of Technology,Kanpur(IIT Kanpur)
 *            {moona,riyaz}@cse.iitk.ac.in
 *
 *   This file is part of IITK OSD Simulator Project.
 *
 *   IITK OSD Simulator Project is free software; you can redistribute it
 *   and/or modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   IITK OSD Simulator Project is distributed in the hope that it will be
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the  GNU General
 *   Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef OSD_OPS_H
#define OSD_OPS_H

#ifndef _KERNEL
#include <stdio.h>
#include <sys/uio.h>
#endif /*ifndef  _KERNEL */

# include "osd.h"
# include "osd_attr2.h"

/**
**********************************
 * File:  osd_ops.h
 * OSD command processing related structures
 * Functions for processing each OSD command
 **/


/** Macros for internal buffers used for */
# define MAX_GET_LIST_HDR_SIZE 512
# define MAX_SET_LIST_HDR_SIZE 512
# define MAX_GET_LIST_DATA_SIZE 512
# define MAX_SET_LIST_DATA_SIZE 512


/**
 *
 *    Function:    osd_data_iov_alloc
 *
 *    Description:  Allocate program OSD data IOV
 * 
 *    Arguments:    @data_buf - OSD data IOV
 *		    @size - size
 *		    @req_iov_size - size of each io vector
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

int osd_data_iov_alloc(OSD_DATA_IOV *data_buf,u32 size,u16 req_iov_size);



/**
 *
 *    Function:    osd_data_iov_free
 *
 *    Description:  Frees program OSD data IOV
 * 
 *    Arguments:    @data_buf - OSD data IOV
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/


int osd_data_iov_free(OSD_DATA_IOV* data_buf);

/**We assume no attribute will span across IO Vector boundary
*/

/** Structure used internally while processing get/set attributes*/
typedef struct 
{
	u32 page_no;
	u32 attr_no;
	u16 attr_len;
	u32 attr_val_offset; /** offset of attribute value in internal 				buffer in  get/set buffer*/
	u32 attr_din_offset;/** offset of attribute value in data in 				buffer */
}ATTR_ENTRY;

typedef ATTR_ENTRY OSD_ATTR_LIST;


/** Structure for OSD Lun */
typedef struct 
{
	char osd_system_id[FILENAME_SIZE];
	uint64_t capacity;
	char username[FILENAME_SIZE];
 	char osd_lun_base_directory[FILENAME_SIZE];
	char osd_target_name[FILENAME_SIZE];
	int lun_no;
}OSD_LUN;

/** Linked list of OSD LUNs in the system*/
typedef struct lun_strcture{
OSD_LUN osd_lun;
struct lun_strcture *next;
}LUN_STRUCT;


/**OSD sense response */
typedef struct 
{
	//u8 status;	
	u8 sense_data_size;
	u8 *sense_buf;	
}OSD_SENSE_RESPONSE;

/** OSD request */
typedef struct 
{
 //char osd_base_directory[FILENAME_SIZE];
 OSD_LUN *osd_lun;
 OSD_CDB_STRUCT osd_cdb_struct;
 char target_name[ISCSI_TARGET_NAME_LEN];
 u32 lun_no;
 u64 part_id;
 u64 object_id;
 int object_type;
 u8 resp_int_chk_value[OSD_ATTR_NUM_CUR_CMD_RESPONSE_INT_CHECK_VAL_size];
 u16 data_iov_size;
 OSD_DATA_IOV data_in; 
 OSD_DATA_IOV data_out; 
 OSD_ATTR_LIST *get_attr_list;
 u8 *get_attr_list_data_buf; 
 u32 get_attr_list_count;
 u32 get_attr_list_data_buf_size;
  OSD_ATTR_LIST *set_attr_list;
 u8 *set_attr_list_data_buf; 
 u32 set_attr_list_count;
 u32 set_attr_list_data_buf_size;
 
 OSD_SENSE_RESPONSE osd_sense_response;
 u8 completed_command_functions[4];
 u8 not_initiated_command_functions[4];
 
 OSD_RESPONSE_HEADER  osd_response_header;
 
}OSD_REQUEST;



/**
 *
 *    Function:    allocate_osd_request
 *
 *    Description:  Allocates new OSD request structure
 *		    Initializes default memebers		
 * 
 *    Arguments:    None
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

OSD_REQUEST *allocate_osd_request();

/**
 *
 *    Function:    free_osd_request
 *
 *    Description:  Frees OSD request 
 *		    
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

int free_osd_request(OSD_REQUEST *);

//int set_osd_req_lun(OSD_REQUEST *osd_req);

/**
 *
 *    Function:    set_osd_req_lun
 *
 *    Description:  Searches for OSD Lun with give OSD Target name in 
 *		    list lun_struct and sets LUN field of OSD request to 
 *		    it
 * 
 *    Arguments:    @osd_req - OSD request
 *		    @lun_struct	- List of all Luns in the system
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/


int set_osd_req_lun(OSD_REQUEST *osd_req,LUN_STRUCT *lun_struct);


/**
 *
 *    Function:    extract_get_attribute_list
		   extract_set_attribute_list	
 *
 *    Description: Extracts get/set attribute list from data in buffer 
 * 	           and stores in get/set attr list structure of osd_req 
 *
 *    Arguments:    @osd_req - OSD Request
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/
int extract_get_attribute_list(OSD_REQUEST *osd_req);
int extract_set_attribute_list(OSD_REQUEST *osd_req);


/**
 *
 *    Function:    process_get_attribute_list
		   process_set_attribute_list	
 *
 *    Description: Process get/set attribute list of osd_req 
 *		   Implement get/set attribute for each attribute in 
 *		   list		
 *
 *    Arguments:   @osd_req - OSD Request
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

int process_get_attribute_list(OSD_REQUEST *osd_req);
int process_set_attribute_list(OSD_REQUEST *osd_req);

/**
 *
 *    Function:    cmnd_opcode
 *
 *    Description: Returns service action for OSD request
 *
 *    Arguments:    @osd_req - OSD Request
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/

u16 cmnd_opcode(OSD_REQUEST *osd_req);


/**
 *
 *    Functions:    osd_xxx(OSD COMMAND NAME)
 *
 *    Description: Implement given OSD command
 *
 *    Arguments:    @osd_req - OSD Request
 *	
 *    Return codes: 0 if success,internal error code otherwise 
 *
 **/
int osd_format(OSD_REQUEST *osd_req);
int osd_create_partition(OSD_REQUEST *osd_req);
int osd_create_collection(OSD_REQUEST *osd_req);
int osd_create(OSD_REQUEST *osd_req);
int osd_flush(OSD_REQUEST *osd_req);
int osd_read(OSD_REQUEST *osd_req);
int osd_write_append(OSD_REQUEST *osd_req);
int osd_create_write(OSD_REQUEST *osd_req);
int osd_list(OSD_REQUEST *osd_req);
int osd_list_collection(OSD_REQUEST *osd_req);
int osd_remove(OSD_REQUEST *osd_req);
int osd_remove_collection(OSD_REQUEST *osd_req);
int osd_remove_partition(OSD_REQUEST *osd_req);
int osd_add_to_collection(OSD_REQUEST *osd_req);
int osd_remove_from_collection(OSD_REQUEST *osd_req);
int osd_get_set_attr(OSD_REQUEST *osd_req);
int osd_other_command(OSD_REQUEST *osd_req);
int osd_perform_task_mgmt(OSD_REQUEST *osd_req);
int osd_scsi_command(OSD_REQUEST *osd_req);

/**
 *
 *    Function:    object_existance_test
 *
 *    Description: Checks whether object with give object type exists or not.
 *		   If not exists,sets sense data for OSD request
 *		   
 *
 *    Arguments:    @osd_req - OSD Request
 *		    @part_id - Partition ID for object
 *		    @obj_id - Object ID for object	
 *		    @object_type - Object Type	
 *
 *    Return codes: 1 if object exists ,0 otherwise 
 *
 **/

int object_existance_test(OSD_REQUEST *osd_req,u64 part_id,u64 obj_id,int object_type); 
# endif /* ifndef OSD_OPS_H */
